import datetime as dt
import logging
from typing import List, Optional

import petname

from cg.constants import DataDelivery, Pipeline
from cg.store import models
from cg.store.api.base import BaseHandler

from cg.constants import Priority

LOG = logging.getLogger(__name__)


class AddHandler(BaseHandler):
    """Methods related to adding new data to the store."""

    def generate_unique_petname(self) -> str:
        while True:
            random_id = petname.Generate(3, separator="")
            if not self.sample(random_id):
                return random_id

    def add_customer(
        self,
        internal_id: str,
        name: str,
        invoice_address: str,
        invoice_reference: str,
        scout_access: bool = False,
        *args,
        **kwargs,
    ) -> models.Customer:
        """Build a new customer record."""

        return self.Customer(
            internal_id=internal_id,
            name=name,
            scout_access=scout_access,
            invoice_address=invoice_address,
            invoice_reference=invoice_reference,
            **kwargs,
        )

    def add_collaboration(self, internal_id: str, name: str, **kwargs) -> models.Collaboration:
        """Build a new customer group record."""

        return self.Collaboration(internal_id=internal_id, name=name, **kwargs)

    def add_user(
        self, customer: models.Customer, email: str, name: str, is_admin: bool = False
    ) -> models.User:
        """Build a new user record."""

        new_user = self.User(name=name, email=email, is_admin=is_admin)
        new_user.customers.append(customer)
        return new_user

    def add_application(
        self,
        tag: str,
        category: str,
        description: str,
        percent_kth: int,
        percent_reads_guaranteed: int,
        is_accredited: bool = False,
        **kwargs,
    ) -> models.Application:
        """Build a new application  record."""

        new_record = self.Application(
            tag=tag,
            prep_category=category,
            description=description,
            is_accredited=is_accredited,
            percent_kth=percent_kth,
            percent_reads_guaranteed=percent_reads_guaranteed,
            **kwargs,
        )
        return new_record

    def add_version(
        self,
        application: models.Application,
        version: int,
        valid_from: dt.datetime,
        prices: dict,
        **kwargs,
    ) -> models.ApplicationVersion:
        """Build a new application version record."""

        new_record = self.ApplicationVersion(version=version, valid_from=valid_from, **kwargs)
        for price_key in [
            Priority.standard.name,
            Priority.priority.name,
            Priority.express.name,
            Priority.research.name,
        ]:
            setattr(new_record, f"price_{price_key}", prices[price_key])
        new_record.application = application
        return new_record

    def add_bed(self, name: str, **kwargs) -> models.Bed:
        """Build a new bed record."""

        new_record = self.Bed(name=name, **kwargs)
        return new_record

    def add_bed_version(
        self, bed: models.Bed, version: int, filename: str, **kwargs
    ) -> models.BedVersion:
        """Build a new bed version record."""

        new_record = self.BedVersion(version=version, filename=filename, **kwargs)
        new_record.bed = bed
        return new_record

    def add_sample(
        self,
        name: str,
        sex: str,
        comment: str = None,
        control: str = None,
        downsampled_to: int = None,
        internal_id: str = None,
        order: str = None,
        ordered: dt.datetime = None,
        priority: Priority = None,
        received: dt.datetime = None,
        ticket: int = None,
        tumour: bool = False,
        **kwargs,
    ) -> models.Sample:
        """Build a new Sample record."""

        internal_id = internal_id or self.generate_unique_petname()
        priority = priority or (Priority.research if downsampled_to else Priority.standard)
        return self.Sample(
            comment=comment,
            control=control,
            downsampled_to=downsampled_to,
            internal_id=internal_id,
            is_tumour=tumour,
            name=name,
            order=order,
            ordered_at=ordered or dt.datetime.now(),
            priority=priority,
            received_at=received,
            sex=sex,
            ticket_number=ticket,
            **kwargs,
        )

    def add_case(
        self,
        data_analysis: Pipeline,
        data_delivery: DataDelivery,
        name: str,
        panels: Optional[List[str]] = None,
        cohorts: Optional[List[str]] = None,
        priority: Optional[Priority] = Priority.standard,
        synopsis: Optional[str] = None,
    ) -> models.Family:
        """Build a new Family record."""

        # generate a unique case id
        while True:
            internal_id = petname.Generate(2, separator="")
            if self.family(internal_id) is None:
                break
            else:
                LOG.debug(f"{internal_id} already used - trying another id")

        new_case = self.Family(
            cohorts=cohorts,
            data_analysis=str(data_analysis),
            data_delivery=str(data_delivery),
            internal_id=internal_id,
            name=name,
            panels=panels,
            priority=priority,
            synopsis=synopsis,
        )
        return new_case

    def relate_sample(
        self,
        family: models.Family,
        sample: models.Sample,
        status: str,
        mother: models.Sample = None,
        father: models.Sample = None,
    ) -> models.FamilySample:
        """Relate a sample record to a family record."""

        new_record = self.FamilySample(status=status)
        new_record.family = family
        new_record.sample = sample
        new_record.mother = mother
        new_record.father = father
        return new_record

    def add_flowcell(
        self, name: str, sequencer: str, sequencer_type: str, date: dt.datetime
    ) -> models.Flowcell:
        """Build a new Flowcell record."""

        new_record = self.Flowcell(
            name=name, sequencer_name=sequencer, sequencer_type=sequencer_type, sequenced_at=date
        )
        return new_record

    def add_analysis(
        self,
        pipeline: Pipeline,
        version: str = None,
        completed_at: dt.datetime = None,
        primary: bool = False,
        uploaded: dt.datetime = None,
        started_at: dt.datetime = None,
        **kwargs,
    ) -> models.Analysis:
        """Build a new Analysis record."""

        new_record = self.Analysis(
            pipeline=str(pipeline),
            pipeline_version=version,
            completed_at=completed_at,
            is_primary=primary,
            uploaded_at=uploaded,
            started_at=started_at,
            **kwargs,
        )
        return new_record

    def add_panel(
        self,
        customer: models.Customer,
        name: str,
        abbrev: str,
        version: float,
        date: dt.datetime = None,
        genes: int = None,
    ) -> models.Panel:
        """Build a new panel record."""

        new_record = self.Panel(
            name=name, abbrev=abbrev, current_version=version, date=date, gene_count=genes
        )
        new_record.customer = customer
        return new_record

    def add_pool(
        self,
        customer: models.Customer,
        name: str,
        order: str,
        ordered: dt.datetime,
        application_version: models.ApplicationVersion,
        ticket: int = None,
        comment: str = None,
        received: dt.datetime = None,
        capture_kit: str = None,
    ) -> models.Pool:
        """Build a new Pool record."""

        new_record = self.Pool(
            name=name,
            ordered_at=ordered or dt.datetime.now(),
            order=order,
            ticket_number=ticket,
            received_at=received,
            comment=comment,
            capture_kit=capture_kit,
        )
        new_record.customer = customer
        new_record.application_version = application_version
        return new_record

    def add_delivery(
        self,
        destination: str,
        sample: models.Sample = None,
        pool: models.Pool = None,
        comment: str = None,
    ) -> models.Delivery:
        """Build a new Delivery record."""

        if not any([sample, pool]):
            raise ValueError("you have to provide a sample or a pool")
        new_record = self.Delivery(destination=destination, comment=comment)
        new_record.sample = sample
        new_record.pool = pool
        return new_record

    def add_invoice(
        self,
        customer: models.Customer,
        samples: List[models.Sample] = None,
        microbial_samples: List[models.Sample] = None,
        pools: List[models.Pool] = None,
        comment: str = None,
        discount: int = 0,
        record_type: str = None,
    ):
        """Build a new Invoice record."""

        new_id = self.new_invoice_id()
        new_invoice = self.Invoice(
            comment=comment, discount=discount, id=new_id, record_type=record_type
        )
        new_invoice.customer = customer
        for sample in samples or []:
            new_invoice.samples.append(sample)
        for microbial_sample in microbial_samples or []:
            new_invoice.samples.append(microbial_sample)
        for pool in pools or []:
            new_invoice.pools.append(pool)
        return new_invoice

    def add_organism(
        self,
        internal_id: str,
        name: str,
        reference_genome: str = None,
        verified: bool = False,
        **kwargs,
    ) -> models.Organism:
        """Build a new Organism record."""

        new_organism = self.Organism(
            internal_id=internal_id,
            name=name,
            reference_genome=reference_genome,
            verified=verified,
            **kwargs,
        )
        return new_organism



from datetime import datetime
import logging
import sys
from pathlib import Path
from typing import TextIO, Optional

import click
from cg.cli.upload.delivery_report import delivery_report_to_scout
from cg.constants import EXIT_SUCCESS, EXIT_FAIL
from cg.exc import CgError
from cg.meta.report.api import ReportAPI
from housekeeper.store import models as hk_models

LOG = logging.getLogger(__name__)

ARGUMENT_CASE_ID = click.argument("case_id", required=True, type=str)
OPTION_STARTED_AT = click.option(
    "--analysis-started-at",
    help="Retrieve analysis started at a specific date (i.e. '2020-05-28  12:00:46')",
)
OPTION_FORCE_REPORT = click.option(
    "-f", "--force", "force_report", is_flag=True, default=False, help="Overrule report validation"
)
OPTION_DRY_RUN = click.option(
    "-d",
    "--dry-run",
    is_flag=True,
    default=False,
    help="Print to console instead of generating an html report file",
)


@click.command("available-delivery-reports")
@OPTION_FORCE_REPORT
@OPTION_DRY_RUN
@click.pass_context
def available_delivery_reports(context: click.Context, force_report: bool, dry_run: bool):
    """Generates delivery reports for all cases that need one and stores them in housekeeper"""

    report_api: ReportAPI = context.obj.meta_apis["report_api"]
    exit_code = EXIT_SUCCESS

    click.echo(click.style("--------------- AVAILABLE DELIVERY REPORTS ---------------"))

    cases_without_delivery_report = report_api.get_cases_without_delivery_report()
    if not cases_without_delivery_report:
        click.echo(
            click.style(
                f"There are no cases available to generate delivery reports ({datetime.now()})",
                fg="green",
            )
        )
    else:
        for analysis_obj in cases_without_delivery_report:
            case_id = analysis_obj.family.internal_id
            LOG.info("Generating delivery report for case: %s", case_id)
            try:
                context.invoke(
                    delivery_report,
                    case_id=analysis_obj.family.internal_id,
                    force_report=force_report,
                    dry_run=dry_run,
                )
            except FileNotFoundError as error:
                LOG.error(
                    "The delivery report generation is missing a file for case: %s, %s",
                    case_id,
                    error,
                )
                exit_code = EXIT_FAIL
            except CgError as error:
                LOG.error(
                    "The delivery report generation failed for case: %s, %s",
                    case_id,
                    error,
                )
                exit_code = EXIT_FAIL
            except Exception as error:
                LOG.error(
                    "Unspecified error when generating the delivery report for case: %s, %s",
                    case_id,
                    error,
                )
                exit_code = EXIT_FAIL

    sys.exit(exit_code)


@click.command("delivery-report")
@ARGUMENT_CASE_ID
@OPTION_FORCE_REPORT
@OPTION_DRY_RUN
@OPTION_STARTED_AT
@click.pass_context
def delivery_report(
    context: click.Context,
    case_id: str,
    force_report: bool,
    dry_run: bool,
    analysis_started_at: str = None,
):
    """Generates a delivery report for a case and stores it in housekeeper"""

    click.echo(click.style("--------------- DELIVERY REPORT ---------------"))

    report_api: ReportAPI = context.obj.meta_apis["report_api"]

    # Missing or not valid internal case ID
    case = report_api.status_db.family(case_id)
    if not case_id or not case:
        LOG.info("Invalid case ID. Retrieving cases without a delivery report.")
        cases_without_delivery_report = report_api.get_cases_without_delivery_report()
        if not cases_without_delivery_report:
            click.echo(
                click.style("There are no cases available to generate delivery reports", fg="green")
            )
        else:
            LOG.error("Provide one of the following case IDs:\n")
            for case_obj in cases_without_delivery_report:
                click.echo(case_obj)

        raise click.Abort

    # Analysis date retrieval
    if not analysis_started_at:
        analysis_started_at: datetime = report_api.status_db.family(case_id).analyses[0].started_at

    # If there is no analysis for the provided date
    if not report_api.status_db.analysis(case, analysis_started_at):
        LOG.error(f"There is no analysis started at {analysis_started_at}")
        raise click.Abort

    LOG.info("Using analysis started at: %s", analysis_started_at)

    # Dry run: prints the HTML report to console
    if dry_run:
        delivery_report_html: str = report_api.create_delivery_report(
            case_id, analysis_started_at, force_report
        )
        click.echo(delivery_report_html)
        return

    delivery_report_file: TextIO = report_api.create_delivery_report_file(
        case_id,
        file_path=Path(report_api.analysis_api.root, case_id),
        analysis_date=analysis_started_at,
        force_report=force_report,
    )

    hk_report_file: Optional[hk_models.File] = report_api.add_delivery_report_to_hk(
        delivery_report_file, case_id, analysis_started_at
    )

    if hk_report_file:
        click.echo(click.style("Uploaded delivery report to housekeeper", fg="green"))
    else:
        click.echo(click.style("Delivery report already uploaded to housekeeper", fg="yellow"))

    context.invoke(delivery_report_to_scout, case_id=case_id, dry_run=dry_run)
    report_api.update_delivery_report_date(case_id=case_id, analysis_date=analysis_started_at)



from pathlib import Path

import pytest


@pytest.fixture(name="yaml_stream")
def yaml_stream() -> str:
    """Return string with yaml format"""
    _content = """- Lorem
- ipsum
- sit
- amet
"""
    return _content


@pytest.fixture(name="invalid_yaml_file")
def invalid_yaml_file(project_dir: Path) -> Path:
    """Return path for invalid yaml file"""
    content = """a: b: c"""
    invalid_yaml: Path = Path(project_dir, "invalid_yaml_file.txt")
    with open(invalid_yaml, "w") as outfile:
        outfile.write(content)
    return invalid_yaml



import json
from datetime import datetime, timedelta
from typing import List

import pytest
from cgmodels.cg.constants import Pipeline
from cg.meta.report.balsamic import BalsamicReportAPI
from cg.meta.report.mip_dna import MipDNAReportAPI
from cg.models.cg_config import CGConfig
from cg.store import models
from tests.apps.scout.conftest import MockScoutApi
from tests.mocks.balsamic_analysis_mock import MockBalsamicAnalysis
from tests.mocks.limsmock import MockLimsAPI
from tests.mocks.mip_analysis_mock import MockMipAnalysis
from tests.mocks.report import MockChanjo, MockDB


@pytest.fixture(scope="function", name="report_api_mip_dna")
def report_api_mip_dna(cg_context: CGConfig, lims_samples) -> MipDNAReportAPI:
    """MIP DNA ReportAPI fixture"""

    cg_context.meta_apis["analysis_api"] = MockMipAnalysis()
    cg_context.status_db_ = MockDB(report_store)
    cg_context.lims_api_ = MockLimsAPI(cg_context, lims_samples)
    cg_context.chanjo_api_ = MockChanjo()
    cg_context.scout_api_ = MockScoutApi(cg_context)
    return MipDNAReportAPI(cg_context, cg_context.meta_apis["analysis_api"])


@pytest.fixture(scope="function", name="report_api_balsamic")
def report_api_balsamic(cg_context: CGConfig, lims_samples) -> BalsamicReportAPI:
    """BALSAMIC ReportAPI fixture"""

    cg_context.meta_apis["analysis_api"] = MockBalsamicAnalysis(cg_context)
    cg_context.status_db_ = MockDB(report_store)
    cg_context.lims_api_ = MockLimsAPI(cg_context, lims_samples)
    cg_context.scout_api_ = MockScoutApi(cg_context)
    return BalsamicReportAPI(cg_context, cg_context.meta_apis["analysis_api"])


@pytest.fixture(scope="function", name="case_mip_dna")
def case_mip_dna(case_id, report_api_mip_dna) -> models.Family:
    """MIP DNA case instance"""

    return report_api_mip_dna.status_db.family(case_id)


@pytest.fixture(scope="function", name="case_balsamic")
def case_balsamic(case_id, report_api_balsamic) -> models.Family:
    """BALSAMIC case instance"""

    return report_api_balsamic.status_db.family(case_id)


@pytest.fixture(scope="function", name="case_samples_data")
def case_samples_data(case_id, report_api_mip_dna):
    """MIP DNA family sample object"""

    return report_api_mip_dna.status_db.family_samples(case_id)


@pytest.fixture(name="mip_analysis_api")
def mip_analysis_api() -> MockMipAnalysis:
    """MIP analysis mock data"""

    return MockMipAnalysis()


@pytest.fixture(name="lims_family")
def lims_family() -> dict:
    """Returns a lims-like case of samples"""

    return json.load(open("tests/fixtures/report/lims_family.json"))


@pytest.fixture(name="lims_samples")
def lims_samples(lims_family: dict) -> List[dict]:
    """Returns the samples of a lims case"""

    return lims_family["samples"]


@pytest.fixture(scope="function", autouse=True, name="report_store")
def report_store(analysis_store, helpers, timestamp_yesterday):
    """A mock store instance for report testing"""

    case = analysis_store.families()[0]
    helpers.add_analysis(
        analysis_store, case, pipeline=Pipeline.MIP_DNA, started_at=timestamp_yesterday
    )
    helpers.add_analysis(analysis_store, case, pipeline=Pipeline.MIP_DNA, started_at=datetime.now())

    # Mock sample dates to calculate processing times
    for family_sample in analysis_store.family_samples(case.internal_id):
        family_sample.sample.ordered_at = timestamp_yesterday - timedelta(days=2)
        family_sample.sample.received_at = timestamp_yesterday - timedelta(days=1)
        family_sample.sample.prepared_at = timestamp_yesterday
        family_sample.sample.sequenced_at = timestamp_yesterday
        family_sample.sample.delivered_at = datetime.now()

    return analysis_store



import logging
from enum import Enum

import cg.store.models
import genologics.entities
from cg.apps.lims import LimsAPI
from cg.store import Store

LOG = logging.getLogger(__name__)


class SampleState(Enum):
    RECEIVED = "received"
    PREPARED = "prepared"
    DELIVERED = "delivered"


class PoolState(Enum):
    RECEIVED = "received"
    DELIVERED = "delivered"


class MicrobialState(Enum):
    RECEIVED = "received"
    PREPARED = "prepared"
    SEQUENCED = "sequenced"
    DELIVERED = "delivered"


class IncludeOptions(Enum):
    UNSET = "unset"
    NOTINVOICED = "not-invoiced"
    ALL = "all"


class TransferLims(object):
    def __init__(self, status: Store, lims: LimsAPI):
        self.status = status
        self.lims = lims

        self._sample_functions = {
            SampleState.RECEIVED: self.status.samples_to_receive,
            SampleState.PREPARED: self.status.samples_to_prepare,
            SampleState.DELIVERED: self.status.samples_to_deliver,
        }

        self._pool_functions = {
            PoolState.RECEIVED: self.status.pools_to_receive,
            PoolState.DELIVERED: self.status.pools_to_deliver,
        }

        self._date_functions = {
            SampleState.RECEIVED: self.lims.get_received_date,
            SampleState.PREPARED: self.lims.get_prepared_date,
            SampleState.DELIVERED: self.lims.get_delivery_date,
            PoolState.RECEIVED: self.lims.get_received_date,
            PoolState.DELIVERED: self.lims.get_delivery_date,
        }

    def _get_all_samples_not_yet_delivered(self):
        return self.status.samples_not_delivered()

    def transfer_samples(
        self, status_type: SampleState, include: str = "unset", sample_id: str = None
    ):
        """Transfer information about samples."""

        if sample_id:
            samples = self.status.Sample.query.filter_by(internal_id=sample_id)
        else:
            samples = self._get_samples_to_include(include, status_type)

        if samples is None:
            LOG.info(f"No samples to process found with {include} {status_type.value}")
            return
        else:
            LOG.info(f"{samples.count()} samples to process")

        for sample_obj in samples:
            lims_date = self._date_functions[status_type](sample_obj.internal_id)
            statusdb_date = getattr(sample_obj, f"{status_type.value}_at")
            if lims_date:

                if statusdb_date and statusdb_date.date() == lims_date:
                    continue

                LOG.info(
                    f"Found new {status_type.value} date for {sample_obj.internal_id}: "
                    f"{lims_date}, old value: {statusdb_date} "
                )

                setattr(sample_obj, f"{status_type.value}_at", lims_date)
                self.status.commit()
            else:
                LOG.debug(f"no {status_type.value} date found for {sample_obj.internal_id}")

    def _get_samples_to_include(self, include, status_type):
        samples = None
        if include == IncludeOptions.UNSET.value:
            samples = self._get_samples_in_step(status_type)
        elif include == IncludeOptions.NOTINVOICED.value:
            samples = self.status.samples_not_invoiced()
        elif include == IncludeOptions.ALL.value:
            samples = self._get_all_relevant_samples()
        return samples

    def transfer_pools(self, status_type: PoolState):
        """Transfer information about pools."""
        pools = self._pool_functions[status_type]()

        for pool_obj in pools:
            ticket_number = pool_obj.ticket_number
            number_of_samples = self.lims.get_sample_number(projectname=ticket_number)
            if not self._is_pool_valid(pool_obj, ticket_number, number_of_samples):
                continue

            samples_in_pool = self.lims.get_samples(projectname=ticket_number)
            for sample_obj in samples_in_pool:
                if not self._is_sample_valid(pool_obj, sample_obj):
                    continue
                status_date = self._date_functions[status_type](sample_obj.id)
                if status_date is None:
                    continue

                LOG.info(
                    "Found %s date for pool id %s: %s",
                    status_type.value,
                    pool_obj.id,
                    status_date,
                )
                setattr(pool_obj, f"{status_type.value}_at", status_date)
                self.status.commit()
                break

    def _get_samples_in_step(self, status_type):
        return self._sample_functions[status_type]()

    def _get_all_relevant_samples(self):
        return self.status.samples_not_downsampled()

    @staticmethod
    def _is_pool_valid(
        pool_obj: cg.store.models.Pool, ticket_number: int, number_of_samples: int
    ) -> bool:
        """Checks if a pool object can be transferred. A pool needs to have a ticket number and at least one sample"""

        if ticket_number is None:
            LOG.warning(f"No ticket number found for pool with order number {pool_obj.order}.")
            return False
        if number_of_samples == 0:
            LOG.warning(f"No samples found for pool with ticket number {ticket_number}.")
            return False
        return True

    @staticmethod
    def _is_sample_valid(
        pool_obj: cg.store.models.Pool, sample_obj: genologics.entities.Sample
    ) -> bool:
        """Checks if a sample can have the status date set. A sample needs to have a udf "pool name" that matches the
        name of the pool object it's part of"""
        if sample_obj.udf.get("pool name") is None:
            LOG.warning(
                "No pool name found for sample %s (sample name %s, project %s)",
                sample_obj.id,
                sample_obj.name,
                sample_obj.project.name,
            )
            return False

        sample_pool_name = sample_obj.udf["pool name"]
        if sample_pool_name != pool_obj.name:
            LOG.warning(
                "The udf 'pool name' of the sample (%s) does not match the name of the pool (%s))",
                sample_pool_name,
                pool_obj.name,
            )
            return False
        return True



"""Fixtures for the meta tests"""

from datetime import datetime

import pytest
from tests.store_helpers import StoreHelpers

from cg.apps.housekeeper.hk import HousekeeperAPI
from cg.constants import Priority
from cg.constants.tags import HkMipAnalysisTag
from cg.meta.workflow.mip_dna import MipDNAAnalysisAPI
from cg.store import Store


@pytest.fixture(scope="function", name="analysis_store")
def fixture_analysis_store(base_store: Store, analysis_family: dict) -> Store:
    """Setup a store instance for testing analysis API."""
    customer = base_store.customer("cust000")
    family = base_store.Family(
        data_analysis=analysis_family["data_analysis"],
        name=analysis_family["name"],
        panels=analysis_family["panels"],
        internal_id=analysis_family["internal_id"],
        priority=Priority.standard.name,
    )
    family.customer = customer
    base_store.add(family)
    application_version = base_store.application("WGSPCFC030").versions[0]
    for sample_data in analysis_family["samples"]:
        sample = base_store.add_sample(
            name=sample_data["name"],
            sex=sample_data["sex"],
            internal_id=sample_data["internal_id"],
            ticket=sample_data["ticket_number"],
            reads=sample_data["reads"],
            capture_kit=sample_data["capture_kit"],
        )
        sample.family = family
        sample.application_version = application_version
        sample.customer = customer
        base_store.add(sample)
    base_store.commit()
    for sample_data in analysis_family["samples"]:
        sample_obj = base_store.sample(sample_data["internal_id"])
        link = base_store.relate_sample(
            family=family,
            sample=sample_obj,
            status=sample_data["status"],
            father=base_store.sample(sample_data["father"]) if sample_data.get("father") else None,
            mother=base_store.sample(sample_data["mother"]) if sample_data.get("mother") else None,
        )
        base_store.add(link)
    base_store.commit()
    yield base_store


@pytest.fixture(scope="function", name="mip_hk_store")
def fixture_mip_hk_store(
    helpers: StoreHelpers,
    real_housekeeper_api: HousekeeperAPI,
    timestamp: datetime,
    case_id: str,
) -> HousekeeperAPI:
    deliver_hk_bundle_data = {
        "name": case_id,
        "created": timestamp,
        "expires": timestamp,
        "files": [
            {
                "path": "tests/fixtures/apps/mip/dna/store/case_config.yaml",
                "archive": False,
                "tags": ["mip-config"],
            },
            {
                "path": "tests/fixtures/apps/mip/dna/store/case_qc_sample_info.yaml",
                "archive": False,
                "tags": ["sampleinfo"],
            },
            {
                "path": "tests/fixtures/apps/mip/case_metrics_deliverables.yaml",
                "archive": False,
                "tags": HkMipAnalysisTag.QC_METRICS,
            },
            {
                "path": "tests/fixtures/apps/mip/case_file.txt",
                "archive": False,
                "tags": ["case-tag"],
            },
            {
                "path": "tests/fixtures/apps/mip/sample_file.txt",
                "archive": False,
                "tags": ["sample-tag", "ADM1"],
            },
        ],
    }
    helpers.ensure_hk_bundle(real_housekeeper_api, deliver_hk_bundle_data, include=True)

    empty_deliver_hk_bundle_data = {
        "name": "case_missing_data",
        "created": timestamp,
        "expires": timestamp,
        "files": [
            {
                "path": "tests/fixtures/apps/mip/dna/store/empty_case_config.yaml",
                "archive": False,
                "tags": ["mip-config"],
            },
            {
                "path": "tests/fixtures/apps/mip/dna/store/empty_case_qc_sample_info.yaml",
                "archive": False,
                "tags": ["sampleinfo"],
            },
            {
                "path": "tests/fixtures/apps/mip/dna/store/empty_case_metrics_deliverables.yaml",
                "archive": False,
                "tags": HkMipAnalysisTag.QC_METRICS,
            },
            {
                "path": "tests/fixtures/apps/mip/case_file.txt",
                "archive": False,
                "tags": ["case-tag"],
            },
            {
                "path": "tests/fixtures/apps/mip/sample_file.txt",
                "archive": False,
                "tags": ["sample-tag", "ADM1"],
            },
        ],
    }
    helpers.ensure_hk_bundle(real_housekeeper_api, empty_deliver_hk_bundle_data, include=True)

    return real_housekeeper_api


@pytest.fixture()
def mip_analysis_api(context_config, mip_hk_store, analysis_store):
    analysis_api = MipDNAAnalysisAPI(context_config)
    analysis_api.housekeeper_api = mip_hk_store
    analysis_api.status_db = analysis_store
    return analysis_api


@pytest.fixture(name="binary_path")
def fixture_binary_path() -> str:
    """Return the string of a path to a (fake) binary"""
    return "/usr/bin/binary"



"""Fixtures for testing apps"""

from pathlib import Path

import pytest
import requests

from cg.utils.commands import Process

# File fixtures for the apps


@pytest.fixture
def balsamic_orderform(orderforms: Path) -> str:
    """Orderform fixture for Balsamic samples"""
    _file = orderforms / "1508.26.balsamic.xlsx"
    return str(_file)


@pytest.fixture
def balsamic_qc_orderform(orderforms: Path) -> str:
    """Orderform fixture for Balsamic QC samples"""
    _file = orderforms / "1508.26.balsamic_qc.xlsx"
    return str(_file)


@pytest.fixture
def balsamic_umi_orderform(orderforms: Path) -> str:
    """Orderform fixture for Balsamic UMI samples"""
    _file = orderforms / "1508.26.balsamic_umi.xlsx"
    return str(_file)


@pytest.fixture
def fastq_orderform(orderforms: Path):
    """Orderform fixture for fastq samples"""
    _file = orderforms / "1508.26.fastq.xlsx"
    return str(_file)


@pytest.fixture
def metagenome_orderform(orderforms: Path) -> str:
    """Orderform fixture for metagenome samples"""
    _file = orderforms / "1605.10.metagenome.xlsx"
    return str(_file)


@pytest.fixture
def mip_orderform(orderforms: Path) -> str:
    """Orderform fixture for MIP samples"""
    _file = orderforms / "1508.26.mip.xlsx"
    return str(_file)


@pytest.fixture
def mip_rna_orderform(orderforms: Path) -> str:
    """Orderform fixture for MIP RNA samples"""
    _file = orderforms / "1508.26.mip_rna.xlsx"
    return str(_file)


@pytest.fixture
def response():
    """Mock a requests.response object"""

    class MockResponse(requests.Response):
        """Mock requests.response class"""

        def __init__(self):
            pass

        @property
        def ok(self):
            """Mock ok"""
            return False

        @property
        def text(self):
            """Mock text"""
            return "response text"

        @property
        def reason(self):
            """Mock reason"""
            return "response reason"

    return MockResponse()


@pytest.fixture
def mock_process():
    """Fixture returns mock Process class factory"""

    def _mock_process(result_stderr: str, result_stdout: str):
        class MockProcess(Process):
            """Process class with mocked run_command method"""

            def run_command(self, parameters=None):
                """Overrides originial run_command method"""
                self.stdout = result_stdout
                self.stderr = result_stderr

        return MockProcess

    return _mock_process



import json
import logging
import os.path
from tempfile import TemporaryDirectory

import requests
from flask import Flask

from cg.exc import TicketCreationError

LOG = logging.getLogger(__name__)
TEXT_FILE_ATTACH_PARAMS = "data:text/plain;charset=utf-8,{content}"


class OsTicket(object):

    """Interface to ticket system."""

    def __init__(self):
        self.headers = None
        self.url = None
        self.osticket_email = None
        self.email_uri = None

    def init_app(self, app: Flask):
        """Initialize the API in Flask."""
        self.setup(
            api_key=app.config["OSTICKET_API_KEY"],
            domain=app.config["OSTICKET_DOMAIN"],
            osticket_email=app.config["SUPPORT_SYSTEM_EMAIL"],
            email_uri=app.config["EMAIL_URI"],
        )

    def setup(
        self,
        api_key: str = None,
        domain: str = None,
        osticket_email: str = None,
        email_uri: str = None,
    ):
        """Initialize the API."""
        self.headers = {"X-API-Key": api_key}
        self.url = os.path.join(domain, "api/tickets.json")
        self.osticket_email = osticket_email
        self.email_uri = email_uri

    def open_ticket(
        self, attachment: dict, email: str, message: str, name: str, subject: str
    ) -> int:
        """Open a new ticket through the REST API."""
        data = dict(
            name=name,
            email=email,
            subject=subject,
            message=message,
            attachments=[attachment],
        )
        res = requests.post(self.url, json=data, headers=self.headers)
        if res.ok:
            return int(res.text)
        LOG.error("res.text: %s, reason: %s", res.text, res.reason)
        raise TicketCreationError(res)

    @staticmethod
    def create_new_ticket_attachment(content: dict, file_name: str) -> dict:
        return {file_name: TEXT_FILE_ATTACH_PARAMS.format(content=json.dumps(content))}

    @staticmethod
    def create_connecting_ticket_attachment(content: dict) -> TemporaryDirectory:
        directory = TemporaryDirectory()
        with open(f"{directory.name}/order.json", "w") as json_file:
            json_file.write(json.dumps(content))
        return directory



"""Fixtures for cli balsamic tests"""
import json
import logging
from datetime import datetime
from pathlib import Path
from tempfile import tempdir

import pytest
from cgmodels.cg.constants import Pipeline
from cg.apps.gt import GenotypeAPI
from cg.apps.housekeeper.hk import HousekeeperAPI
from cg.apps.scout.scoutapi import ScoutAPI
from cg.apps.tb import TrailblazerAPI
from cg.constants.delivery import MIP_DNA_ANALYSIS_CASE_TAGS, PIPELINE_ANALYSIS_TAG_MAP
from cg.constants.tags import HkMipAnalysisTag, HK_DELIVERY_REPORT_TAG
from cg.meta.deliver import DeliverAPI
from cg.meta.report.mip_dna import MipDNAReportAPI
from cg.meta.rsync import RsyncAPI
from cg.meta.upload.scout.uploadscoutapi import UploadScoutAPI
from cg.meta.workflow.mip import MipAnalysisAPI
from cg.meta.workflow.mip_dna import MipDNAAnalysisAPI
from cg.models.cg_config import CGConfig
from cg.models.scout.scout_load_config import ScoutLoadConfig
from cg.store import Store, models
from tests.meta.upload.scout.conftest import fixture_mip_load_config
from tests.mocks.hk_mock import MockHousekeeperAPI
from tests.mocks.madeline import MockMadelineAPI

LOG = logging.getLogger(__name__)


@pytest.fixture(name="scout_load_config")
def fixture_scout_load_config(apps_dir: Path) -> Path:
    """Yaml file with load information from scout"""
    return apps_dir / "scout/643594.config.yaml"


@pytest.fixture(scope="function", name="scout_hk_bundle_data")
def fixture_scout_hk_bundle_data(case_id: str, scout_load_config: Path, timestamp: datetime):
    """Get some bundle data for housekeeper"""
    tag_name = UploadScoutAPI.get_load_config_tag()

    return {
        "name": case_id,
        "created": timestamp,
        "expires": timestamp,
        "files": [
            {
                "path": str(scout_load_config),
                "archive": False,
                "tags": [tag_name],
            }
        ],
    }


@pytest.fixture(name="upload_genotypes_hk_bundle")
def fixture_upload_genotypes_hk_bundle(
    case_id: str, timestamp, case_qc_metrics_deliverables: Path, bcf_file: Path
) -> dict:
    """Returns a dictionary in hk format with files used in upload gt process"""
    return {
        "name": case_id,
        "created": datetime.now(),
        "expires": datetime.now(),
        "files": [
            {
                "path": str(case_qc_metrics_deliverables),
                "archive": False,
                "tags": HkMipAnalysisTag.QC_METRICS,
            },
            {"path": str(bcf_file), "archive": False, "tags": ["snv-gbcf", "genotype"]},
        ],
    }


@pytest.fixture(name="analysis_obj")
def fixture_analysis_obj(
    analysis_store_trio: Store, case_id: str, timestamp: datetime, helpers
) -> models.Analysis:
    """Return a analysis object with a trio"""
    return analysis_store_trio.family(case_id).analyses[0]


@pytest.fixture(name="upload_genotypes_hk_api")
def fixture_upload_genotypes_hk_api(
    real_housekeeper_api: HousekeeperAPI,
    upload_genotypes_hk_bundle: dict,
    analysis_obj: models.Analysis,
    helpers,
) -> HousekeeperAPI:
    """Add and include files from upload genotypes hk bundle"""
    helpers.ensure_hk_bundle(real_housekeeper_api, upload_genotypes_hk_bundle)
    hk_version = real_housekeeper_api.last_version(analysis_obj.family.internal_id)
    real_housekeeper_api.include(hk_version)
    return real_housekeeper_api


@pytest.fixture(name="upload_genotypes_context")
def fixture_upload_genotypes_context(
    upload_genotypes_hk_api: HousekeeperAPI,
    genotype_api: GenotypeAPI,
    analysis_store_trio: Store,
    base_context: CGConfig,
) -> CGConfig:
    """Create a upload genotypes context"""
    base_context.genotype_api_ = genotype_api
    base_context.housekeeper_api_ = upload_genotypes_hk_api
    base_context.status_db_ = analysis_store_trio
    return base_context


@pytest.fixture(name="upload_report_hk_bundle")
def fixture_upload_report_hk_bundle(case_id: str, delivery_report_html: Path, timestamp) -> dict:
    """Returns a dictionary including the delivery report html file"""

    return {
        "name": case_id,
        "created": datetime.now(),
        "files": [
            {"path": str(delivery_report_html), "archive": False, "tags": [HK_DELIVERY_REPORT_TAG]}
        ],
    }


@pytest.fixture(name="upload_report_hk_api")
def fixture_upload_report_hk_api(
    real_housekeeper_api: HousekeeperAPI,
    upload_report_hk_bundle: dict,
    analysis_obj: models.Analysis,
    helpers,
) -> HousekeeperAPI:
    """Add and include files from upload reports hk bundle"""

    helpers.ensure_hk_bundle(real_housekeeper_api, upload_report_hk_bundle)
    hk_version = real_housekeeper_api.last_version(analysis_obj.family.internal_id)
    real_housekeeper_api.include(hk_version)
    return real_housekeeper_api


@pytest.fixture(name="scout_load_object")
def fixture_scout_load_object(case_id: str, timestamp: datetime) -> ScoutLoadConfig:
    """Create a scout load config case object"""
    case_data = {
        "owner": "cust000",
        "case": case_id,
        "human_genome_build": "37",
        "rank_score_threshold": 5,
        "analysis_date": timestamp,
        "samples": [
            {"sample_id": "sample", "sex": "male", "phenotype": "affected", "analysis_type": "wgs"}
        ],
    }
    return ScoutLoadConfig(**case_data)


@pytest.fixture(name="base_context")
def fixture_base_context(
    analysis_store: Store,
    housekeeper_api: HousekeeperAPI,
    upload_scout_api: UploadScoutAPI,
    trailblazer_api: TrailblazerAPI,
    cg_context: CGConfig,
) -> CGConfig:
    """context to use in cli"""
    cg_context.status_db_ = analysis_store
    cg_context.housekeeper_api_ = housekeeper_api
    cg_context.trailblazer_api_ = trailblazer_api
    cg_context.scout_api_ = MockScoutApi()
    cg_context.meta_apis["scout_upload_api"] = upload_scout_api
    cg_context.mip_rd_dna.root = tempdir

    return cg_context


@pytest.fixture(name="fastq_context")
def fixture_fastq_context(
    base_context,
    analysis_store: Store,
    housekeeper_api: HousekeeperAPI,
    upload_scout_api: UploadScoutAPI,
    trailblazer_api: TrailblazerAPI,
    cg_context: CGConfig,
) -> CGConfig:
    """Fastq context to use in cli"""

    base_context.meta_apis["delivery_api"] = DeliverAPI(
        store=base_context.status_db,
        hk_api=base_context.housekeeper_api,
        case_tags=PIPELINE_ANALYSIS_TAG_MAP[Pipeline.FASTQ]["case_tags"],
        sample_tags=PIPELINE_ANALYSIS_TAG_MAP[Pipeline.FASTQ]["sample_tags"],
        delivery_type="fastq",
        project_base_path=Path(base_context.delivery_path),
    )
    base_context.meta_apis["rsync_api"] = RsyncAPI(cg_context)
    base_context.trailblazer_api_ = trailblazer_api
    return base_context


@pytest.fixture(scope="function", name="upload_scout_api")
def fixture_upload_scout_api(housekeeper_api: MockHousekeeperAPI, mip_load_config: ScoutLoadConfig):
    """Return a upload scout api"""
    api = MockScoutUploadApi()
    api.housekeeper = housekeeper_api
    api.config = mip_load_config

    return api


class MockScoutApi(ScoutAPI):
    def __init__(self):
        """docstring for __init__"""

    def upload(self, scout_load_config: Path, threshold: int = 5, force: bool = False):
        """docstring for upload"""
        LOG.info("Case loaded successfully to Scout")


class MockAnalysisApi(MipAnalysisAPI):
    def __init__(self):
        """docstring for __init__"""

    def get_latest_metadata(self, internal_id):
        """docstring for upload"""
        return {}


class MockScoutUploadApi(UploadScoutAPI):
    def __init__(self, **kwargs):
        """docstring for __init__"""
        self.mock_generate_config = True
        self.housekeeper = None
        self.madeline_api = MockMadelineAPI()
        self.analysis = MockAnalysisApi()
        self.config = ScoutLoadConfig()
        self.file_exists = False
        self.lims = MockLims()
        self.missing_mandatory_field = False

    @pytest.fixture(autouse=True)
    def _request_analysis(self, analysis_store_single_case):
        self.analysis = analysis_store_single_case

    def generate_config(self, analysis_obj, **kwargs):
        """Mock the generate config"""
        if self.missing_mandatory_field:
            self.config.vcf_snv = None

        return self.config

    def save_config_file(self, upload_config, file_path):
        """docstring for save_config_file"""
        return

    def add_scout_config_to_hk(self, config_file_path: str, case_id: str, delete: bool = False):
        """docstring for add_scout_config_to_hk"""
        LOG.info("Use mock to upload file")
        if self.file_exists:
            raise FileExistsError("Scout config already exists")


class MockLims:
    """Mock lims fixture"""

    lims = None

    def __init__(self):
        self.lims = self

    @staticmethod
    def lims_samples():
        """Return LIMS-like case samples"""
        lims_family = json.load(open("tests/fixtures/report/lims_family.json"))
        return lims_family["samples"]

    def sample(self, sample_id):
        """Returns a lims sample matching the provided sample_id"""
        for sample in self.lims_samples():
            if sample["id"] == sample_id:
                return sample
        return None


@pytest.fixture(name="upload_context")
def upload_context(cg_context: CGConfig) -> CGConfig:
    analysis_api = MipDNAAnalysisAPI(config=cg_context)
    cg_context.meta_apis["analysis_api"] = analysis_api
    cg_context.meta_apis["report_api"] = MipDNAReportAPI(cg_context, analysis_api)
    cg_context.meta_apis["scout_upload_api"] = UploadScoutAPI(
        hk_api=cg_context.housekeeper_api,
        scout_api=cg_context.scout_api,
        madeline_api=cg_context.madeline_api,
        analysis_api=analysis_api,
        lims_api=cg_context.lims_api,
        status_db=cg_context.status_db,
    )

    return cg_context



"""This script tests the cli methods to set flowcells in status-db"""
from cg.cli.set.base import flowcell
from cg.constants import FLOWCELL_STATUS
from cg.models.cg_config import CGConfig
from cg.store import Store
from click.testing import CliRunner

SUCCESS = 0


def test_set_flowcell_bad_flowcell(cli_runner: CliRunner, base_context: CGConfig):
    """Test to set a flowcell using a non-existing flowcell"""
    # GIVEN an empty database

    # WHEN setting a flowcell
    flowcell_name = "dummy_name"
    result = cli_runner.invoke(flowcell, [flowcell_name], obj=base_context)

    # THEN it should complain in missing flowcell instead of setting a flowcell
    assert result.exit_code != SUCCESS


def test_set_flowcell_required(
    cli_runner: CliRunner, base_context: CGConfig, base_store: Store, helpers
):
    """Test to set a flowcell using only the required arguments"""
    # GIVEN a database with a flowcell
    flowcell_name = helpers.add_flowcell(base_store).name
    assert base_store.Flowcell.query.count() == 1

    # WHEN setting a flowcell
    result = cli_runner.invoke(flowcell, [flowcell_name], obj=base_context)

    # THEN it should have been set
    assert result.exit_code == SUCCESS


def test_set_flowcell_status(
    cli_runner: CliRunner, base_context: CGConfig, base_store: Store, helpers
):
    """Test that the updated flowcell get the status we send in"""
    # GIVEN a database with a flowcell
    flowcell_name = helpers.add_flowcell(base_store).name
    status = FLOWCELL_STATUS[2]
    assert base_store.Flowcell.query.first().status != status

    # WHEN setting a flowcell
    result = cli_runner.invoke(flowcell, ["--status", status, flowcell_name], obj=base_context)

    # THEN it should have been set
    assert result.exit_code == SUCCESS
    assert base_store.Flowcell.query.count() == 1
    assert base_store.Flowcell.query.first().status == status



"""Upload API"""

import logging
from datetime import datetime, timedelta

import click
from cg.exc import AnalysisUploadError, AnalysisAlreadyUploadedError
from cg.meta.workflow.analysis import AnalysisAPI
from cg.meta.upload.scout.uploadscoutapi import UploadScoutAPI
from cg.models.cg_config import CGConfig
from cg.meta.meta import MetaAPI
from cg.store import models


LOG = logging.getLogger(__name__)


class UploadAPI(MetaAPI):
    """Upload API"""

    def __init__(self, config: CGConfig, analysis_api: AnalysisAPI):
        super().__init__(config=config)
        self.analysis_api = analysis_api
        self.scout_upload_api: UploadScoutAPI = UploadScoutAPI(
            hk_api=config.housekeeper_api,
            scout_api=config.scout_api,
            madeline_api=config.madeline_api,
            analysis_api=self.analysis_api,
            lims_api=config.lims_api,
            status_db=config.status_db,
        )

    def upload(self, ctx: click.Context, case_obj: models.Family, restart: bool) -> None:
        """Uploads pipeline specific analysis data and files"""

        raise NotImplementedError

    def update_upload_started_at(self, analysis: models.Analysis) -> None:
        """Updates the upload_started_at field with the current local date and time"""

        analysis.upload_started_at = datetime.now()
        self.status_db.commit()

    def update_uploaded_at(self, analysis: models.Analysis) -> None:
        """Updates the uploaded_at field with the current local date and time"""

        analysis.uploaded_at = datetime.now()
        self.status_db.commit()

    @staticmethod
    def verify_analysis_upload(case_obj: models.Family, restart: bool) -> None:
        """Verifies the state of an analysis upload in StatusDB"""

        if not case_obj.data_delivery:
            LOG.error(f"The data delivery field has not been specified for: {case_obj.internal_id}")
            raise AnalysisUploadError

        if not case_obj.analyses:
            LOG.error(f"There is no analysis for case: {case_obj.internal_id}")
            raise AnalysisUploadError

        analysis_obj: models.Analysis = case_obj.analyses[0]
        if analysis_obj.uploaded_at is not None:
            LOG.error(f"The analysis has been already uploaded: {analysis_obj.uploaded_at.date()}")
            raise AnalysisAlreadyUploadedError

        if not restart and analysis_obj.upload_started_at is not None:
            if datetime.now() - analysis_obj.upload_started_at > timedelta(hours=24):
                LOG.error(
                    f"This upload has already started at {analysis_obj.upload_started_at}, but something went wrong. "
                    f"Restart it with the --restart flag."
                )
                raise AnalysisUploadError

            LOG.warning(f"The upload has already started: {analysis_obj.upload_started_at.time()}")
            raise AnalysisAlreadyUploadedError



import logging
from typing import List

import click

from cg.meta.report.balsamic import BalsamicReportAPI
from cg.meta.report.balsamic_umi import BalsamicUmiReportAPI
from cg.meta.workflow.balsamic import BalsamicAnalysisAPI
from cg.meta.workflow.balsamic_umi import BalsamicUmiAnalysisAPI
from cg.meta.workflow.mip_dna import MipDNAAnalysisAPI
from cg.meta.report.mip_dna import MipDNAReportAPI
from cg.models.cg_config import CGConfig
from cg.meta.report.api import ReportAPI
from housekeeper.store import models as hk_models

LOG = logging.getLogger(__name__)


@click.group("mip-dna")
@click.pass_context
def mip_dna(context: click.Context):
    """Upload MIP-DNA files to Scout"""

    context.obj.meta_apis["report_api"] = MipDNAReportAPI(
        config=context.obj, analysis_api=MipDNAAnalysisAPI(config=context.obj)
    )


@click.group("balsamic")
@click.pass_context
def balsamic(context: click.Context):
    """Upload BALSAMIC files to Scout"""

    context.obj.meta_apis["report_api"] = BalsamicReportAPI(
        config=context.obj, analysis_api=BalsamicAnalysisAPI(config=context.obj)
    )


@click.group("balsamic-umi")
@click.pass_context
def balsamic_umi(context: click.Context):
    """Upload BALSAMIC UMI files to Scout"""

    context.obj.meta_apis["report_api"] = BalsamicUmiReportAPI(
        config=context.obj, analysis_api=BalsamicUmiAnalysisAPI(config=context.obj)
    )


@click.command("delivery-report-to-scout")
@click.argument("case_id", required=False, type=str)
@click.option(
    "-d", "--dry-run", is_flag=True, default=False, help="Run command without uploading to scout"
)
@click.pass_obj
def delivery_report_to_scout(context: CGConfig, case_id: str, dry_run: bool):
    """Fetches a delivery report from housekeeper and uploads it to scout"""

    report_api: ReportAPI = context.meta_apis["report_api"]

    # Missing case ID
    if not case_id:
        LOG.info("Case ID not provided. Retrieving cases with pending reports to upload to Scout.")
        cases_without_delivery_report = report_api.get_cases_without_delivery_report()
        if not cases_without_delivery_report:
            click.echo(click.style("There are no reports to upload to Scout", fg="green"))
        else:
            LOG.error("Provide one of the following case IDs:\n")
            for case_obj in cases_without_delivery_report:
                click.echo(case_obj)

        raise click.Abort

    uploaded_delivery_report_files: List[hk_models.File] = [
        file_obj
        for file_obj in report_api.housekeeper_api.get_files(
            bundle=case_id,
            tags=["delivery-report"],
            version=report_api.housekeeper_api.last_version(case_id).id,
        )
    ]

    if not uploaded_delivery_report_files:
        LOG.error(f"No delivery report was found in housekeeper for case: {case_id}")
        raise FileNotFoundError

    report_path: str = uploaded_delivery_report_files[0].full_path

    if not dry_run:
        report_api.scout_api.upload_delivery_report(
            report_path=report_path, case_id=case_id, update=True
        )


mip_dna.add_command(delivery_report_to_scout)
balsamic.add_command(delivery_report_to_scout)
balsamic_umi.add_command(delivery_report_to_scout)



"""Commands to generates and/or modify BALSAMIC files"""

import click

from cg.meta.report.balsamic import BalsamicReportAPI
from cg.meta.report.balsamic_umi import BalsamicUmiReportAPI
from cg.meta.workflow.balsamic import BalsamicAnalysisAPI
from cg.cli.generate.commands import delivery_report, available_delivery_reports
from cg.meta.workflow.balsamic_umi import BalsamicUmiAnalysisAPI


@click.group("balsamic", invoke_without_command=True)
@click.pass_context
def balsamic(context: click.Context):
    """Balsamic file generation/modification"""

    if context.invoked_subcommand is None:
        click.echo(context.get_help())
        return

    context.obj.meta_apis["report_api"] = BalsamicReportAPI(
        config=context.obj, analysis_api=BalsamicAnalysisAPI(config=context.obj)
    )


@click.group("balsamic-umi", invoke_without_command=True)
@click.pass_context
def balsamic_umi(context: click.Context):
    """Balsamic file generation/modification"""

    if context.invoked_subcommand is None:
        click.echo(context.get_help())
        return

    context.obj.meta_apis["report_api"] = BalsamicUmiReportAPI(
        config=context.obj, analysis_api=BalsamicUmiAnalysisAPI(config=context.obj)
    )


balsamic.add_command(delivery_report)
balsamic.add_command(available_delivery_reports)
balsamic_umi.add_command(delivery_report)
balsamic_umi.add_command(available_delivery_reports)



import logging
from pathlib import Path

import click

from cg.apps.demultiplex.demultiplex_api import DemultiplexingAPI
from cg.apps.tb import TrailblazerAPI
from cg.constants.demultiplexing import OPTION_BCL_CONVERTER
from cg.exc import FlowcellError
from cg.meta.demultiplex.delete_demultiplex_api import DeleteDemuxAPI
from cg.models.cg_config import CGConfig
from cg.models.demultiplex.flowcell import Flowcell

LOG = logging.getLogger(__name__)

DRY_RUN = click.option("--dry-run", is_flag=True)


@click.command(name="all")
@OPTION_BCL_CONVERTER
@DRY_RUN
@click.option("--flowcells-directory", type=click.Path(exists=True, file_okay=False))
@click.pass_obj
def demultiplex_all(
    context: CGConfig, bcl_converter: str, flowcells_directory: click.Path, dry_run: bool
):
    """Demultiplex all flowcells that are ready under the flowcells_directory"""
    LOG.info("Running cg demultiplex all, using %s.", bcl_converter)
    if flowcells_directory:
        flowcells_directory: Path = Path(str(flowcells_directory))
    else:
        flowcells_directory: Path = Path(context.demultiplex.run_dir)
    demultiplex_api: DemultiplexingAPI = context.demultiplex_api
    demultiplex_api.set_dry_run(dry_run=dry_run)
    tb_api: TrailblazerAPI = context.trailblazer_api
    LOG.info("Search for flowcells ready to demultiplex in %s", flowcells_directory)
    for sub_dir in flowcells_directory.iterdir():
        if not sub_dir.is_dir():
            continue
        LOG.info("Found directory %s", sub_dir)
        try:
            flowcell_obj = Flowcell(flowcell_path=sub_dir, bcl_converter=bcl_converter)
        except FlowcellError:
            continue

        if not demultiplex_api.is_demultiplexing_possible(flowcell=flowcell_obj) and not dry_run:
            continue

        if not flowcell_obj.validate_sample_sheet():
            LOG.warning(
                "Malformed sample sheet. Run cg demultiplex samplesheet validate %s",
                flowcell_obj.sample_sheet_path,
            )
            continue

        delete_demux_api: DeleteDemuxAPI = DeleteDemuxAPI(
            config=context,
            demultiplex_base=demultiplex_api.out_dir,
            dry_run=dry_run,
            run_path=(flowcells_directory / sub_dir),
        )

        delete_demux_api.delete_flow_cell(
            cg_stats=False,
            demultiplexing_dir=True,
            run_dir=False,
            housekeeper=True,
            init_files=False,
            status_db=False,
        )

        slurm_job_id: int = demultiplex_api.start_demultiplexing(flowcell=flowcell_obj)
        demultiplex_api.add_to_trailblazer(
            tb_api=tb_api, slurm_job_id=slurm_job_id, flowcell=flowcell_obj
        )


@click.command(name="flowcell")
@click.argument("flowcell-id")
@DRY_RUN
@OPTION_BCL_CONVERTER
@click.pass_obj
def demultiplex_flowcell(
    context: CGConfig,
    dry_run: bool,
    flowcell_id: str,
    bcl_converter: str,
):
    """Demultiplex a flowcell on slurm using CG

    flowcell-id is the flowcell run directory name, e.g. '201203_A00689_0200_AHVKJCDRXX'
    """

    LOG.info("Running cg demultiplex flowcell, using %s.", bcl_converter)
    flowcell_directory: Path = Path(context.demultiplex.run_dir) / flowcell_id

    demultiplex_api: DemultiplexingAPI = context.demultiplex_api
    demultiplex_api.set_dry_run(dry_run=dry_run)
    LOG.info(f"SETTING FLOWCELL ID TO {flowcell_id}")
    LOG.info(f"SETTING OUT DIR TO {demultiplex_api.out_dir}")

    try:
        flowcell_obj = Flowcell(flowcell_path=flowcell_directory, bcl_converter=bcl_converter)
    except FlowcellError as e:
        raise click.Abort from e

    delete_demux_api: DeleteDemuxAPI = DeleteDemuxAPI(
        config=context,
        demultiplex_base=demultiplex_api.out_dir,
        dry_run=dry_run,
        run_path=flowcell_directory,
    )

    delete_demux_api.delete_flow_cell(
        cg_stats=True,
        demultiplexing_dir=True,
        run_dir=False,
        housekeeper=True,
        init_files=True,
        status_db=False,
    )

    if not demultiplex_api.is_demultiplexing_possible(flowcell=flowcell_obj) and not dry_run:
        LOG.warning("Can not start demultiplexing!")
        return

    if not flowcell_obj.validate_sample_sheet():
        LOG.warning(
            "Malformed sample sheet. Run cg demultiplex samplesheet validate %s",
            flowcell_obj.sample_sheet_path,
        )
        raise click.Abort

    slurm_job_id: int = demultiplex_api.start_demultiplexing(flowcell=flowcell_obj)
    tb_api: TrailblazerAPI = context.trailblazer_api
    demultiplex_api.add_to_trailblazer(
        tb_api=tb_api, slurm_job_id=slurm_job_id, flowcell=flowcell_obj
    )


@click.command(name="delete-flow-cell")
@click.option("--cg-stats", is_flag=True, help="Delete flow cell in cg-stats")
@click.option(
    "-d",
    "--demultiplex-base",
    type=click.Path(exists=True),
    required=True,
    help="Is the base of demultiplexing, e.g. '/home/proj/{ENVIRONMENT}/demultiplexed-runs/'",
)
@click.option(
    "--demultiplexing-dir", is_flag=True, help="Delete flow cell demultiplexed dir on file system"
)
@DRY_RUN
@click.option("--housekeeper", is_flag=True, help="Delete flow cell in housekeeper")
@click.option("--init-files", is_flag=True, help="Delete flow cell init-files")
@click.option("--run-dir", is_flag=True, help="Delete flow cell run on file system")
@click.option(
    "-r",
    "--run-directory",
    type=click.Path(exists=False),
    required=True,
    help="Is the path to the flowcell run directory name, e.g. '/home/proj/{ENVIRONMENT}/flowcells/novaseq/runs/201203_A00689_0200_AHVKJCDRXX'",
)
@click.option(
    "--status-db",
    is_flag=True,
    help="Delete flow cell in status-db, if passed all other entries are also deleted",
)
@click.option("--yes", is_flag=True, help="Pass yes to click confirm")
@click.pass_obj
def delete_flow_cell(
    context: CGConfig,
    dry_run: bool,
    demultiplexing_dir: str,
    cg_stats: bool,
    demultiplex_base: bool,
    housekeeper: bool,
    init_files: bool,
    run_dir: bool,
    run_directory: str,
    status_db: bool,
    yes: bool,
):
    """Delete a flowcell. If --status-db is passed then all other options will be treated as True"""
    demultiplex_base: Path = Path(demultiplex_base)
    run_path: Path = Path(run_directory)

    wipe_demux_api: DeleteDemuxAPI = DeleteDemuxAPI(
        config=context, demultiplex_base=demultiplex_base, dry_run=dry_run, run_path=run_path
    )

    if yes or click.confirm(
        f"Are you sure you want to delete the flow cell from the following databases:\n"
        f"cg-stats={True if status_db else cg_stats}\nDemultiplexing-dir={True if status_db else demultiplexing_dir}\n"
        f"Housekeeper={True if status_db else housekeeper}\nInit_files={True if status_db else init_files}\n"
        f"Run-dir={True if status_db else run_dir}\nStatusdb={status_db}"
    ):
        wipe_demux_api.delete_flow_cell(
            cg_stats=cg_stats,
            demultiplexing_dir=demultiplexing_dir,
            housekeeper=housekeeper,
            init_files=init_files,
            run_dir=run_dir,
            status_db=status_db,
        )



from typing import Optional

from sqlalchemy import Column, ForeignKey, UniqueConstraint, orm, types
from sqlalchemy.orm.exc import NoResultFound

from .base import Model


class Demux(Model):

    __table_args__ = (UniqueConstraint("flowcell_id", "basemask", name="demux_ibuk_1"),)

    demux_id = Column(types.Integer, primary_key=True)
    flowcell_id = Column(ForeignKey("flowcell.flowcell_id", ondelete="CASCADE"), nullable=False)
    datasource_id = Column(
        ForeignKey("datasource.datasource_id", ondelete="CASCADE"), nullable=False
    )
    basemask = Column(types.String(255))
    time = Column(types.DateTime)

    datasource = orm.relationship("Datasource", cascade="all", backref=orm.backref("demuxes"))
    unaligned = orm.relation(
        "Unaligned", single_parent=True, cascade="all, delete-orphan", backref=orm.backref("demux")
    )
    # add the viewonly attribute to make sure the Session.delete(demux) works
    samples = orm.relation(
        "Sample",
        secondary="unaligned",
        viewonly=True,
        cascade="all",
        backref=orm.backref("demuxes"),
    )

    @staticmethod
    def exists(flowcell_id: int, basemask: str) -> Optional[int]:
        """Checks if the Demux entry already exists"""
        try:
            demux: Demux = (
                Demux.query.filter_by(flowcell_id=flowcell_id).filter_by(basemask=basemask).one()
            )
            return demux.demux_id
        except NoResultFound:
            return None



""" This file groups all tests related to microsalt start creation """
import logging

from cg.cli.workflow.microsalt.base import run
from cg.models.cg_config import CGConfig
from click.testing import CliRunner

EXIT_SUCCESS = 0


def test_no_arguments(cli_runner: CliRunner, base_context: CGConfig):
    """Test command without any options"""

    # GIVEN

    # WHEN dry running without anything specified
    result = cli_runner.invoke(run, obj=base_context)

    # THEN command should mention missing arguments
    assert result.exit_code != EXIT_SUCCESS


def test_dry_arguments(cli_runner: CliRunner, base_context: CGConfig, microbial_ticket, caplog):
    """Test command dry"""

    # GIVEN
    caplog.set_level(logging.INFO)

    # WHEN dry running without anything specified
    result = cli_runner.invoke(run, [microbial_ticket, "-t", "-d"], obj=base_context)

    # THEN command should mention missing arguments
    assert result.exit_code == EXIT_SUCCESS
    assert f"Running command" in caplog.text



import click
import datetime as dt
import logging
import shutil

from pathlib import Path

from cgmodels.cg.constants import Pipeline
from cg.apps.housekeeper.hk import HousekeeperAPI
from cg.constants import EXIT_FAIL, EXIT_SUCCESS
from cg.exc import FlowcellsNeededError, DecompressionNeededError
from cg.meta.rsync import RsyncAPI
from cg.meta.workflow.analysis import AnalysisAPI
from cg.meta.workflow.balsamic import BalsamicAnalysisAPI
from cg.meta.workflow.fluffy import FluffyAnalysisAPI
from cg.meta.workflow.mip_dna import MipDNAAnalysisAPI
from cg.meta.workflow.mutant import MutantAnalysisAPI
from cg.models.cg_config import CGConfig
from cg.store import Store, models
from dateutil.parser import parse as parse_date

OPTION_DRY = click.option(
    "-d", "--dry-run", help="Simulate process without executing", is_flag=True
)
OPTION_YES = click.option("-y", "--yes", is_flag=True, help="Skip confirmation")
ARGUMENT_BEFORE_STR = click.argument("before_str", type=str)
ARGUMENT_CASE_ID = click.argument("case_id", required=True)
OPTION_ANALYSIS_PARAMETERS_CONFIG = click.option(
    "--config-artic", type=str, help="Config with computational and lab related settings"
)

LOG = logging.getLogger(__name__)


@click.command("ensure-flowcells-ondisk")
@ARGUMENT_CASE_ID
@click.pass_obj
def ensure_flowcells_ondisk(context: CGConfig, case_id: str):
    """Check if flowcells are on disk for given case. If not, request flowcells and raise FlowcellsNeededError"""
    analysis_api: AnalysisAPI = context.meta_apis["analysis_api"]
    analysis_api.verify_case_id_in_statusdb(case_id=case_id)
    if not analysis_api.all_flowcells_on_disk(case_id=case_id):
        raise FlowcellsNeededError(
            "Analysis cannot be started: all flowcells need to be on disk to run the analysis"
        )
    LOG.info("All flowcells present on disk")


@click.command("resolve-compression")
@ARGUMENT_CASE_ID
@OPTION_DRY
@click.pass_obj
def resolve_compression(context: CGConfig, case_id: str, dry_run: bool):
    """Handles cases where decompression is needed before starting analysis"""
    analysis_api: AnalysisAPI = context.meta_apis["analysis_api"]
    analysis_api.verify_case_id_in_statusdb(case_id=case_id)
    is_decompression_running: bool = analysis_api.resolve_decompression(
        case_id=case_id, dry_run=dry_run
    )
    if is_decompression_running:
        raise DecompressionNeededError("Workflow interrupted: decompression is not finished")


@click.command("link")
@ARGUMENT_CASE_ID
@OPTION_DRY
@click.pass_obj
def link(context: CGConfig, case_id: str, dry_run: bool):
    """Link FASTQ files for all samples in a case"""
    analysis_api: AnalysisAPI = context.meta_apis["analysis_api"]
    analysis_api.verify_case_id_in_statusdb(case_id)
    if dry_run:
        return
    analysis_api.link_fastq_files(case_id=case_id)


@click.command("store")
@ARGUMENT_CASE_ID
@OPTION_DRY
@click.pass_obj
def store(context: CGConfig, case_id: str, dry_run: bool):
    """Store finished analysis files in Housekeeper"""

    analysis_api: AnalysisAPI = context.meta_apis["analysis_api"]
    housekeeper_api: HousekeeperAPI = context.housekeeper_api
    status_db: Store = context.status_db
    analysis_api.verify_case_id_in_statusdb(case_id=case_id)

    if dry_run:
        LOG.info("Dry run: Would have stored deliverables for %s", case_id)
        return
    try:
        analysis_api.upload_bundle_housekeeper(case_id=case_id)
        analysis_api.upload_bundle_statusdb(case_id=case_id)
        analysis_api.set_statusdb_action(case_id=case_id, action=None)
    except Exception as exception_object:
        housekeeper_api.rollback()
        status_db.rollback()
        LOG.error("Error storing deliverables for case %s - %s", case_id, exception_object)
        raise


@click.command("store-available")
@OPTION_DRY
@click.pass_context
def store_available(context: click.Context, dry_run: bool) -> None:
    """Store bundles for all finished analyses in Housekeeper"""

    analysis_api: AnalysisAPI = context.obj.meta_apis["analysis_api"]

    exit_code: int = EXIT_SUCCESS
    for case_obj in analysis_api.get_cases_to_store():
        LOG.info("Storing deliverables for %s", case_obj.internal_id)
        try:
            context.invoke(store, case_id=case_obj.internal_id, dry_run=dry_run)
        except Exception as exception_object:
            LOG.error("Error storing %s: %s", case_obj.internal_id, exception_object)
            exit_code = EXIT_FAIL
    if exit_code:
        raise click.Abort


@click.command("rsync-past-run-dirs")
@OPTION_YES
@OPTION_DRY
@ARGUMENT_BEFORE_STR
@click.pass_obj
def rsync_past_run_dirs(context: CGConfig, before_str: str, dry_run: bool, yes: bool) -> None:
    """Remove deliver workflow commands"""

    rsync_api: RsyncAPI = RsyncAPI(config=context)

    before: dt.datetime = parse_date(before_str)

    for process in rsync_api.rsync_processes:
        if rsync_api.process_ready_to_clean(before=before, process=process):
            if yes or click.confirm(f"Do you want to remove all files in {process}?"):
                if dry_run:
                    LOG.info(f"Would have removed {process}")
                    continue
                LOG.info(f"Removing {process.as_posix()}")
                shutil.rmtree(process, ignore_errors=True)
        else:
            LOG.info(f"{process.as_posix()} is still young")


@click.command("clean-run-dir")
@OPTION_YES
@OPTION_DRY
@ARGUMENT_CASE_ID
@click.pass_obj
def clean_run_dir(context: CGConfig, yes: bool, case_id: str, dry_run: bool = False):
    """Remove workflow run directory"""

    analysis_api: AnalysisAPI = context.meta_apis["analysis_api"]
    status_db: Store = context.status_db
    analysis_api.verify_case_id_in_statusdb(case_id)
    analysis_api.verify_case_path_exists(case_id=case_id)
    analysis_api.check_analysis_ongoing(case_id=case_id)
    analysis_path: Path = analysis_api.get_case_path(case_id)

    if dry_run:
        LOG.info(f"Would have deleted: {analysis_path}")
        return EXIT_SUCCESS

    if yes or click.confirm(f"Are you sure you want to remove all files in {analysis_path}?"):
        if analysis_path.is_symlink():
            LOG.warning(
                f"Will not automatically delete symlink: {analysis_path}, delete it manually",
            )
            return EXIT_FAIL

        shutil.rmtree(analysis_path, ignore_errors=True)
        LOG.info("Cleaned %s", analysis_path)
        analyses: list = status_db.family(case_id).analyses
        for analysis_obj in analyses:
            analysis_obj.cleaned_at = analysis_obj.cleaned_at or dt.datetime.now()
            status_db.commit()


@click.command("past-run-dirs")
@OPTION_YES
@OPTION_DRY
@ARGUMENT_BEFORE_STR
@click.pass_context
def past_run_dirs(
    context: click.Context, before_str: str, yes: bool = False, dry_run: bool = False
):
    """Clean up of old case run dirs"""

    exit_code = EXIT_SUCCESS
    analysis_api: AnalysisAPI = context.obj.meta_apis["analysis_api"]

    before = parse_date(before_str)
    possible_cleanups = analysis_api.get_analyses_to_clean(before=before)
    LOG.info(f"Cleaning {len(possible_cleanups)} analyses created before {before}")

    for analysis in possible_cleanups:
        case_id = analysis.family.internal_id
        try:
            LOG.info("Cleaning %s output for %s", analysis_api.pipeline, case_id)
            context.invoke(clean_run_dir, yes=yes, case_id=case_id, dry_run=dry_run)
        except Exception as error:
            LOG.error("Failed to clean directories for case %s - %s", case_id, error)
            exit_code = EXIT_FAIL
    if exit_code:
        raise click.Abort
    LOG.info("Done cleaning %s output ", analysis_api.pipeline)


@click.command("balsamic-past-run-dirs")
@OPTION_YES
@OPTION_DRY
@ARGUMENT_BEFORE_STR
@click.pass_context
def balsamic_past_run_dirs(
    context: click.Context, before_str: str, yes: bool = False, dry_run: bool = False
):
    """Clean up of "old" Balsamic case run dirs"""

    context.obj.meta_apis["analysis_api"] = BalsamicAnalysisAPI(context.obj)

    context.invoke(past_run_dirs, yes=yes, dry_run=dry_run, before_str=before_str)


@click.command("fluffy-past-run-dirs")
@OPTION_YES
@OPTION_DRY
@ARGUMENT_BEFORE_STR
@click.pass_context
def fluffy_past_run_dirs(
    context: click.Context, before_str: str, yes: bool = False, dry_run: bool = False
):
    """Clean up of "old" Fluffy case run dirs"""

    context.obj.meta_apis["analysis_api"] = FluffyAnalysisAPI(context.obj)

    context.invoke(past_run_dirs, yes=yes, dry_run=dry_run, before_str=before_str)


@click.command("mip-past-run-dirs")
@OPTION_YES
@OPTION_DRY
@ARGUMENT_BEFORE_STR
@click.pass_context
def mip_past_run_dirs(
    context: click.Context, before_str: str, yes: bool = False, dry_run: bool = False
):
    """Clean up of "old" MIP case run dirs"""

    context.obj.meta_apis["analysis_api"] = MipDNAAnalysisAPI(context.obj)

    context.invoke(past_run_dirs, yes=yes, dry_run=dry_run, before_str=before_str)


@click.command("mutant-past-run-dirs")
@OPTION_YES
@OPTION_DRY
@ARGUMENT_BEFORE_STR
@click.pass_context
def mutant_past_run_dirs(
    context: click.Context, before_str: str, yes: bool = False, dry_run: bool = False
):
    """Clean up of "old" MUTANT case run dirs"""

    context.obj.meta_apis["analysis_api"] = MutantAnalysisAPI(context.obj)

    context.invoke(past_run_dirs, yes=yes, dry_run=dry_run, before_str=before_str)



"""
This module handles concatenation of balsamic fastq files.

Classes:
    FastqFileNameCreator: Creates valid balsamic filenames
    FastqHandler: Handles fastq file linking
"""
import datetime as dt
import gzip
import logging
import os
import re
import shutil
from pathlib import Path
from typing import List, Optional

from pydantic.main import BaseModel

LOG = logging.getLogger(__name__)

DEFAULT_DATE_STR = (
    "171015"  # Stand in value to use if sequencing date cannot be extracted from header
)
DEFAULT_INDEX = (
    "XXXXXX"  # Stand in value to use if flowcell index is to be masked when renaming file
)


class FastqHandler:
    """Handles fastq file linking"""

    @staticmethod
    def concatenate(files: List, concat_file: str):
        """Concatenates a list of fastq files"""
        LOG.info(FastqHandler.display_files(files, concat_file))

        with open(concat_file, "wb") as write_file_obj:
            for filename in files:
                with open(filename, "rb") as file_descriptor:
                    shutil.copyfileobj(file_descriptor, write_file_obj)

        size_before = FastqHandler.size_before(files)
        size_after = FastqHandler.size_after(concat_file)

        try:
            FastqHandler.assert_file_sizes(size_before, size_after)
        except AssertionError as error:
            LOG.warning(error)

    @staticmethod
    def size_before(files: List) -> int:
        """Returns the total size of the linked fastq files before concatenation"""

        return sum(os.stat(f).st_size for f in files)

    @staticmethod
    def size_after(concat_file: str) -> int:
        """returns the size of the concatenated fastq files"""

        return os.stat(concat_file).st_size

    @staticmethod
    def assert_file_sizes(size_before: int, size_after: int) -> None:
        """asserts the file sizes before and after concatenation. If the file sizes differ by more
        than 1 percent, throw an exception"""
        msg = (
            f"Warning: Large file size difference after concatenation!"
            f"Before: {size_before} -> after: {size_after}"
        )

        assert abs(size_before - size_after) / size_before <= 0.01, msg
        LOG.info("Concatenation file size check successful!")

    @staticmethod
    def display_files(files: List, concat_file: str) -> str:
        """display file names for logging purposes"""
        return (
            f"Concatenating: {', '.join(Path(file_).name for file_ in files)} -> "
            f"{Path(concat_file).name}"
        )

    @staticmethod
    def remove_files(files: list) -> None:
        for file_name in files:
            file_path: Path = Path(file_name)
            if not file_path.is_file():
                continue
            file_path.unlink()

    @staticmethod
    def get_concatenated_name(linked_fastq_name: str) -> str:
        """ "create a name for the concatenated file for some read files"""
        return f"concatenated_{'_'.join(linked_fastq_name.split('_')[-4:])}"

    @staticmethod
    def parse_header(line: str) -> dict:
        """Generates a dict with parsed lanes, flowcells and read numbers
        Handle illumina's two different header formats
        @see https://en.wikipedia.org/wiki/FASTQ_format

        @HWUSI-EAS100R:6:73:941:1973#0/1

            HWUSI-EAS100R   the unique instrument name
            6   flowcell lane
            73  tile number within the flowcell lane
            941     'x'-coordinate of the cluster within the tile
            1973    'y'-coordinate of the cluster within the tile
            #0  index number for a multiplexed sample (0 for no indexing)
            /1  the member of a pair, /1 or /2 (paired-end or mate-pair reads only)

        Versions of the Illumina pipeline since 1.4 appear to use #NNNNNN
        instead of #0 for the multiplex ID, where NNNNNN is the sequence of the
        multiplex tag.

        With Casava 1.8 the format of the '@' line has changed:

        @EAS139:136:FC706VJ:2:2104:15343:197393 1:Y:18:ATCACG

            EAS139  the unique instrument name
            136     the run id
            FC706VJ     the flowcell id
            2   flowcell lane
            2104    tile number within the flowcell lane
            15343   'x'-coordinate of the cluster within the tile
            197393  'y'-coordinate of the cluster within the tile
            1   the member of a pair, 1 or 2 (paired-end or mate-pair reads only)
            Y   Y if the read is filtered, N otherwise
            18  0 when none of the control bits are on, otherwise it is an even number
            ATCACG  index sequence
        """

        fastq_meta = {"lane": None, "flowcell": None, "readnumber": None}

        parts = line.split(":")
        if len(parts) == 5:  # @HWUSI-EAS100R:6:73:941:1973#0/1
            fastq_meta["lane"] = parts[1]
            fastq_meta["flowcell"] = "XXXXXX"
            fastq_meta["readnumber"] = parts[-1].split("/")[-1]
        if len(parts) == 10:  # @EAS139:136:FC706VJ:2:2104:15343:197393 1:Y:18:ATCACG
            fastq_meta["lane"] = parts[3]
            fastq_meta["flowcell"] = parts[2]
            fastq_meta["readnumber"] = parts[6].split(" ")[-1]
        if len(parts) == 7:  # @ST-E00201:173:HCLCGALXX:1:2106:22516:34834/1
            fastq_meta["lane"] = parts[3]
            fastq_meta["flowcell"] = parts[2]
            fastq_meta["readnumber"] = parts[-1].split("/")[-1]

        return fastq_meta

    @staticmethod
    def parse_file_data(fastq_path: Path) -> dict:
        with gzip.open(fastq_path) as handle:
            header_line = handle.readline().decode()
            header_info = FastqHandler.parse_header(header_line)

            data = {
                "path": fastq_path,
                "lane": int(header_info["lane"]),
                "flowcell": header_info["flowcell"],
                "read": int(header_info["readnumber"]),
                "undetermined": ("Undetermined" in fastq_path),
            }
            matches = re.findall(r"-l[1-9]t([1-9]{2})_", str(fastq_path))
            if len(matches) > 0:
                data["flowcell"] = f"{data['flowcell']}-{matches[0]}"
            return data

    @staticmethod
    def create_fastq_name(
        lane: str,
        flowcell: str,
        sample: str,
        read: str,
        date: dt.datetime = DEFAULT_DATE_STR,
        index: str = DEFAULT_INDEX,
        undetermined: Optional[str] = None,
        meta: Optional[str] = None,
    ):
        raise NotImplementedError


class BalsamicFastqHandler(FastqHandler):
    @staticmethod
    def create_fastq_name(
        lane: str,
        flowcell: str,
        sample: str,
        read: str,
        date: dt.datetime = DEFAULT_DATE_STR,
        index: str = DEFAULT_INDEX,
        undetermined: Optional[str] = None,
        meta: Optional[str] = None,
    ) -> str:
        """Name a FASTQ file following Balsamic conventions. Naming must be
        xxx_R_1.fastq.gz and xxx_R_2.fastq.gz"""
        flowcell = f"{flowcell}-undetermined" if undetermined else flowcell
        date_str = date if isinstance(date, str) else date.strftime("%y%m%d")
        return f"{lane}_{date_str}_{flowcell}_{sample}_{index}_R_{read}.fastq.gz"


class MipFastqHandler(FastqHandler):
    @staticmethod
    def create_fastq_name(
        lane: str,
        flowcell: str,
        sample: str,
        read: str,
        date: dt.datetime = DEFAULT_DATE_STR,
        index: str = DEFAULT_INDEX,
        undetermined: Optional[str] = None,
        meta: Optional[str] = None,
    ) -> str:
        """Name a FASTQ file following MIP conventions."""
        flowcell = f"{flowcell}-undetermined" if undetermined else flowcell
        date_str = date if isinstance(date, str) else date.strftime("%y%m%d")
        return f"{lane}_{date_str}_{flowcell}_{sample}_{index}_{read}.fastq.gz"


class MicrosaltFastqHandler(FastqHandler):
    @staticmethod
    def create_fastq_name(
        lane: str,
        flowcell: str,
        sample: str,
        read: str,
        date: dt.datetime = DEFAULT_DATE_STR,
        index: str = DEFAULT_INDEX,
        undetermined: Optional[str] = None,
        meta: Optional[str] = None,
    ) -> str:
        """Name a FASTQ file following usalt conventions. Naming must be
        xxx_R_1.fastq.gz and xxx_R_2.fastq.gz"""
        # ACC1234A1_FCAB1ABC2_L1_1.fastq.gz sample_flowcell_lane_read.fastq.gz

        flowcell = f"{flowcell}-undetermined" if undetermined else flowcell
        return f"{sample}_{flowcell}_L{lane}_{read}.fastq.gz"


class MutantFastqHandler(FastqHandler):
    @staticmethod
    def create_fastq_name(
        lane: str,
        flowcell: str,
        sample: str,
        read: str,
        date: dt.datetime = DEFAULT_DATE_STR,
        index: str = DEFAULT_INDEX,
        undetermined: Optional[str] = None,
        meta: Optional[str] = None,
    ) -> str:
        """Name a FASTQ file following mutant conventions. Naming must be
        xxx_R_1.fastq.gz and xxx_R_2.fastq.gz"""
        # ACC1234A1_FCAB1ABC2_L1_1.fastq.gz sample_flowcell_lane_read.fastq.gz

        return f"{flowcell}_L{lane}_{meta}_{read}.fastq.gz"

    @staticmethod
    def get_concatenated_name(linked_fastq_name: str) -> str:
        """ "Create a name for the concatenated file from multiple lanes"""
        return "_".join(linked_fastq_name.split("_")[2:])

    @staticmethod
    def get_nanopore_header_info(line: str) -> dict:
        fastq_meta = {"flowcell": None}
        header_metadata: list = line.split(" ")
        flowcell = header_metadata[5].split("=")
        fastq_meta["flowcell"] = flowcell[1]
        return fastq_meta

    @staticmethod
    def create_nanopore_fastq_name(
        flowcell: str,
        sample: str,
        filenr: str,
        meta: Optional[str] = None,
    ) -> str:
        return f"{flowcell}_{sample}_{meta}_{filenr}.fastq.gz"

    @staticmethod
    def parse_nanopore_file_data(fastq_path: Path) -> dict:
        with gzip.open(fastq_path) as handle:
            header_line = handle.readline().decode()
            header_info: dict = MutantFastqHandler.get_nanopore_header_info(line=header_line)
            data = {
                "path": fastq_path,
                "flowcell": header_info["flowcell"],
            }
            return data



import logging

import click
from cg.cli.demultiplex.add import add_flowcell_cmd, select_project_cmd
from cg.cli.demultiplex.demux import demultiplex_all, demultiplex_flowcell, delete_flow_cell
from cg.cli.demultiplex.finish import finish_group
from cg.cli.demultiplex.report import create_report_cmd
from cg.cli.demultiplex.sample_sheet import sample_sheet_commands

LOG = logging.getLogger(__name__)


@click.group(name="demultiplex")
def demultiplex_cmd_group():
    """Command group for the demultiplex commands"""
    LOG.info("Running CG demultiplex")


demultiplex_cmd_group: click.Group
demultiplex_cmd_group.add_command(demultiplex_flowcell)
demultiplex_cmd_group.add_command(demultiplex_all)
demultiplex_cmd_group.add_command(sample_sheet_commands)
demultiplex_cmd_group.add_command(add_flowcell_cmd)
demultiplex_cmd_group.add_command(select_project_cmd)
demultiplex_cmd_group.add_command(finish_group)
demultiplex_cmd_group.add_command(create_report_cmd)
demultiplex_cmd_group.add_command(delete_flow_cell)



from flask_admin import Admin
from flask_alchy import Alchy
from flask_cors import CORS

from cg.apps.lims import LimsAPI
from cg.apps.osticket import OsTicket
from cg.store import api, models


class CgAlchy(Alchy, api.CoreHandler):
    pass


class FlaskLims(LimsAPI):
    def __init__(self, app=None):
        if app:
            self.init_app(app)

    def init_app(self, app):
        config = {
            "lims": {
                "host": app.config["LIMS_HOST"],
                "username": app.config["LIMS_USERNAME"],
                "password": app.config["LIMS_PASSWORD"],
            }
        }
        super(FlaskLims, self).__init__(config)


cors = CORS(resources={r"/api/*": {"origins": "*"}}, supports_credentials=True)
db = CgAlchy(Model=models.Model)
admin = Admin(name="Clinical Genomics")
lims = FlaskLims()
osticket = OsTicket()



"""Fixtures for the scout api tests"""

from pathlib import Path

import pytest

from cg.constants.pedigree import Pedigree
from cg.constants.subject import RelationshipStatus, Gender, PhenotypeStatus
from tests.mocks.process_mock import ProcessMock

from cg.apps.scout.scoutapi import ScoutAPI


class MockScoutApi(ScoutAPI):
    def __init__(self, config: dict):
        binary_path = "scout"
        config_path = "config_path"
        self.process = ProcessMock(binary=binary_path, config=config_path)


@pytest.fixture(name="sample_dict")
def fixture_sample_dict() -> dict:
    sample_dict = {
        "analysis_type": "wgs",
        "bam_path": "/path/to/sample.bam",
        "mt_bam": "/path/to/reduced_mt.bam",
        "capture_kit": None,
        Pedigree.FATHER: RelationshipStatus.HAS_NO_PARENT,
        Pedigree.MOTHER: RelationshipStatus.HAS_NO_PARENT,
        "sample_id": "sample_id",
        "sample_name": "sample_name",
        Pedigree.SEX: Gender.MALE,
        "tissue_type": "unknown",
        Pedigree.PHENOTYPE: PhenotypeStatus.AFFECTED,
    }
    return dict(sample_dict)


@pytest.fixture(name="omim_disease_nr")
def fixture_omim_disease_nr() -> int:
    return 607208


@pytest.fixture(name="scout_dir")
def fixture_scout_dir(apps_dir: Path) -> Path:
    """Return the path to the scout fixtures dir"""
    return Path(apps_dir, "scout")


@pytest.fixture(name="causatives_file")
def fixture_causatives_file(scout_dir: Path) -> Path:
    """Return the path to a file with causatives output"""
    return Path(scout_dir, "export_causatives.json")


@pytest.fixture(name="cases_file")
def fixture_cases_file(scout_dir: Path) -> Path:
    """Return the path to a file with export cases output"""
    return Path(scout_dir, "case_export.json")


@pytest.fixture(name="none_case_file")
def fixture_none_case_file(scout_dir: Path) -> Path:
    """Return the path to a file with export cases output where mandatory fields are None"""
    return Path(scout_dir, "none_case_export.json")


@pytest.fixture(name="other_sex_case_file")
def fixture_other_sex_case_file(scout_dir: Path) -> Path:
    """Return the path to a file with export cases output where sex is 'other0'"""
    return Path(scout_dir, "other_sex_case.json")


@pytest.fixture(name="none_case_output")
def fixture_none_case_output(none_case_file: Path) -> str:
    """Return the content of a export causatives run with scout"""
    with open(none_case_file, "r") as infile:
        content = infile.read()
    return content


@pytest.fixture(name="other_sex_case_output")
def fixture_other_sex_case_output(other_sex_case_file: Path) -> str:
    """Return the content of a export causatives run with scout"""
    with open(other_sex_case_file, "r") as infile:
        content = infile.read()
    return content


@pytest.fixture(name="causative_output")
def fixture_causative_output(causatives_file: Path) -> str:
    """Return the content of a export causatives run with scout"""
    with open(causatives_file, "r") as infile:
        content = infile.read()
    return content


@pytest.fixture(name="export_cases_output")
def fixture_export_cases_output(cases_file: Path) -> str:
    """Return the content of a export cases output"""
    with open(cases_file, "r") as infile:
        content = infile.read()
    return content


@pytest.fixture(name="scout_api")
def fixture_scout_api() -> ScoutAPI:
    return MockScoutApi({})



"""MIP-DNA upload API"""

import logging

import click
from cg.cli.upload.scout import scout
from cg.cli.upload.observations import observations
from cg.cli.upload.genotype import genotypes
from cg.cli.upload.validate import validate
from cg.cli.upload.coverage import coverage
from cg.constants import DataDelivery
from cg.meta.workflow.mip_dna import MipDNAAnalysisAPI
from cg.models.cg_config import CGConfig
from cg.meta.upload.upload_api import UploadAPI
from cg.store import models


LOG = logging.getLogger(__name__)


class MipDNAUploadAPI(UploadAPI):
    """MIP-DNA upload API"""

    def __init__(self, config: CGConfig):
        self.analysis_api: MipDNAAnalysisAPI = MipDNAAnalysisAPI(config)
        super().__init__(config=config, analysis_api=self.analysis_api)

    def upload(self, ctx: click.Context, case_obj: models.Family, restart: bool) -> None:
        """Uploads MIP-DNA analysis data and files"""

        analysis_obj: models.Analysis = case_obj.analyses[0]
        self.update_upload_started_at(analysis_obj)

        # Main upload
        ctx.invoke(coverage, family_id=case_obj.internal_id, re_upload=restart)
        ctx.invoke(validate, family_id=case_obj.internal_id)
        ctx.invoke(genotypes, family_id=case_obj.internal_id, re_upload=restart)
        ctx.invoke(observations, case_id=case_obj.internal_id)

        # Scout specific upload
        if DataDelivery.SCOUT in case_obj.data_delivery:
            ctx.invoke(scout, case_id=case_obj.internal_id, re_upload=restart)

        self.update_uploaded_at(analysis_obj)



import itertools
import logging
import shutil

from glob import glob
from pathlib import Path
from typing import Iterable, List, Optional

from cg.apps.housekeeper.hk import HousekeeperAPI
from cg.apps.cgstats.stats import StatsAPI
from cg.exc import DeleteDemuxError
from cg.models.cg_config import CGConfig
from cg.store import Store
from cg.store.models import Sample, Flowcell
from housekeeper.store.models import File

log = logging.getLogger(__name__)


class DeleteDemuxAPI:
    """Class to handle wiping out a flow cell before restart/start"""

    def __init__(self, config: CGConfig, demultiplex_base: Path, dry_run: bool, run_path: Path):
        self.dry_run: bool = self.set_dry_run(dry_run=dry_run)
        self.demultiplexing_dir: Path = demultiplex_base
        self.housekeeper_api: HousekeeperAPI = config.housekeeper_api
        self.run_path: Path = run_path
        self.status_db: Store = config.status_db
        self.stats_api: StatsAPI = config.cg_stats_api
        self.samples_on_flow_cell: List[Sample] = []
        log.debug("DeleteDemuxAPI: API initiated")

    @property
    def demultiplexing_path(self) -> Path:
        """Return the demultiplexing directory path for the given run name"""
        return self.demultiplexing_dir.joinpath(self.run_path.name)

    @property
    def flow_cell_name(self) -> str:
        """Parse flow cell name from flowcell run name

        This will assume that the flow cell naming convention is used.
        Convention is: <date>_<machine>_<run_numbers>_<A|B><flowcell_id>
        Example: 201203_A00689_0200_AHVKJCDRXX
        """
        return self.run_path.name.split("_")[-1][1:]

    @property
    def status_db_presence(self) -> bool:
        """Update about the presence of given flow cell in status_db"""
        return bool(
            self.status_db.query(Flowcell).filter(Flowcell.name == self.flow_cell_name).first()
        )

    @staticmethod
    def set_dry_run(dry_run: bool) -> bool:
        """Set dry run flag for API"""
        log.debug(f"DeleteDemuxAPI: Setting dry run mode to {dry_run}")
        return dry_run

    def _set_samples_on_flow_cell(self) -> None:
        """Set a list of samples related to a flow cell in status-db"""
        self.samples_on_flow_cell: List[Sample] = (
            self.status_db.query(Flowcell)
            .filter(Flowcell.name == self.flow_cell_name)
            .first()
            .samples
        )

    def active_samples_on_flow_cell(self) -> Optional[List[str]]:
        """Check if there are any active cases related to samples of a flow cell"""
        active_samples_on_flow_cell: List[str] = [
            sample.internal_id
            for sample in self.samples_on_flow_cell
            if self.status_db.active_sample(internal_id=sample.internal_id)
        ]
        if active_samples_on_flow_cell:
            return active_samples_on_flow_cell

    def _delete_sample_sheet_housekeeper(self) -> None:
        """Delete the presence of all sample sheets related to a flow cell in Housekeeper."""
        sample_sheet_files: Iterable[File] = self.housekeeper_api.files(
            tags=[self.flow_cell_name, "samplesheet"]
        )
        if any(sample_sheet_files):
            for file in sample_sheet_files:
                self.housekeeper_api.delete_file(file_id=file.id)
                log.info(f"DeleteDemuxAPI-Housekeeper: Deleted {file.path} from housekeeper")
        else:
            log.info(f"DeleteDemuxAPI-Housekeeper: No files found with tag: {self.flow_cell_name}")

    def _delete_files_if_related_in_housekeeper_by_tag(self, sample: Sample, tag: str):
        """Delete any existing fastq related to sample"""

        housekeeper_files: Iterable[File] = self.housekeeper_api.files(
            bundle=sample.internal_id, tags=[tag]
        )
        if not housekeeper_files:
            log.info(f"Could not find {tag} for {sample.internal_id}")
        else:
            for housekeeper_file in housekeeper_files:
                self.housekeeper_api.delete_file_if_related(
                    stem=self.demultiplexing_path.as_posix(), hk_file=housekeeper_file
                )

    def _delete_fastq_and_spring_housekeeper(self) -> None:
        """Delete the presence of any spring/fastq files in Housekeeper related to samples on the flow cell"""

        tags = ["fastq", "spring"]
        for tag, sample in itertools.product(tags, self.samples_on_flow_cell):
            self._delete_files_if_related_in_housekeeper_by_tag(sample=sample, tag=tag)

    def delete_flow_cell_housekeeper(self) -> None:
        """Delete any presence of a flow cell in housekeeper. Including Sample sheets AND fastq-files"""

        if self.dry_run:
            log.info(
                f"DeleteDemuxAPI-Housekeeper: Would delete sample sheet files with tag {self.flow_cell_name}"
            )
            log.info(
                f"DeleteDemuxAPI-Housekeeper: Would delete fastq and spring files related to flowcell {self.flow_cell_name}"
            )
        else:
            log.info(
                f"DeleteDemuxAPI-Housekeeper: Deleting sample sheet files with tag {self.flow_cell_name}"
            )
            self._delete_sample_sheet_housekeeper()
            log.info(
                f"DeleteDemuxAPI-Housekeeper: Deleting fastq and spring files related to flowcell {self.flow_cell_name}"
            )
            self._delete_fastq_and_spring_housekeeper()

    def delete_flow_cell_statusdb(self) -> None:
        """Delete any presence of a flow cell in status-db"""
        if self.dry_run:
            log.info(f"DeleteDemuxAPI-StatusDB: Would remove {self.flow_cell_name}")
        else:
            self.status_db.delete_flowcell(flowcell_name=self.flow_cell_name)
            log.info(f"DeleteDemuxAPI-StatusDB: Deleted flowcell {self.flow_cell_name}")

    def delete_flow_cell_cgstats(self) -> None:
        """Delete any presence of a flow cell in cgstats"""
        from cg.apps.cgstats.crud.delete import delete_flowcell

        if self.dry_run:
            log.info(f"DeleteDemuxAPI-CGStats: Would remove {self.flow_cell_name}")
        else:
            delete_flowcell(manager=self.stats_api, flowcell_name=self.flow_cell_name)

    def _delete_demultiplexing_dir_hasta(self) -> None:
        """delete demultiplexing directory on server"""
        log.info(
            f"DeleteDemuxAPI-Hasta: Removing flow cell demultiplexing directory: {self.demultiplexing_path}"
        )
        shutil.rmtree(self.demultiplexing_path, ignore_errors=False)

    def _delete_run_dir_hasta(self) -> None:
        """delete flow cell run directory on server"""
        log.info(f"DeleteDemuxAPI-Hasta: Removing flow cell run directory: {self.run_path}")
        shutil.rmtree(path=self.run_path, ignore_errors=False)

    def delete_flow_cell_hasta(
        self,
        demultiplexing_dir: bool,
        run_dir: bool,
    ) -> None:
        """Delete a flow cells presence on the server, if flow cell is removed from demultiplexing dir and run
        dir set status to removed"""
        if self.dry_run:
            log.info(
                f"DeleteDemuxAPI-Hasta: Would have removed the following directory: {self.demultiplexing_path}\n"
                f"DeleteDemuxAPI-Hasta: Would have removed the following directory: {self.run_path}"
            )
            return
        if demultiplexing_dir and run_dir and self.status_db_presence:
            flow_cell_obj: Flowcell = self.status_db.flowcell(self.flow_cell_name)
            flow_cell_obj.status = "removed"
            self.status_db.commit()
        if demultiplexing_dir and self.demultiplexing_path.exists():
            self._delete_demultiplexing_dir_hasta()
        else:
            log.info(
                f"DeleteDemuxAPI-Hasta: Skipped demultiplexing directory, or no target: {self.demultiplexing_dir}"
            )
        if run_dir and self.run_path.exists():
            self._delete_run_dir_hasta()
        else:
            log.info(
                f"DeleteDemuxAPI-Hasta: Skipped flow cell run directory, or no target: {self.run_path}"
            )

    def delete_demux_init_files(self):
        """Delete previous traces of slurm job ids"""
        slurm_job_id_file_path: Path = self.run_path / "slurm_job_ids.yaml"
        demux_script_file_path: Path = self.run_path / "demux-novaseq.sh"
        try:
            error_log_path, log_path = glob(
                f"{self.run_path}/{self.flow_cell_name}_demultiplex.std*"
            )
            demux_init_files: List[Path] = [
                slurm_job_id_file_path,
                demux_script_file_path,
                Path(error_log_path),
                Path(log_path),
            ]
        except ValueError:
            log.info(f"DeleteDemuxAPI-Init-files: No demultiplexing logs found in: {self.run_path}")
            demux_init_files: List[Path] = [
                slurm_job_id_file_path,
                demux_script_file_path,
            ]
        if self.dry_run:
            log.info(f"DeleteDemuxAPI-Init-files: Would have removed {demux_init_files}")
        for init_file in demux_init_files:
            if init_file.is_file():
                log.info(f"DeleteDemuxAPI: Removing {init_file}")
                init_file.unlink()
            else:
                log.info(f"DeleteDemuxAPI: Did not find {init_file}")

    def check_active_samples(self) -> None:
        """Check, and raise, if there are active samples on a flow cell"""
        if self.status_db_presence:
            self._set_samples_on_flow_cell()
            active_samples_on_flow_cell: List[str] = self.active_samples_on_flow_cell()
            if active_samples_on_flow_cell:
                log.warning(
                    f"There are active analyses using data from this flowcell.\n"
                    f"Before restarting the demultiplexing process - please cancel the ongoing analyses of the "
                    f"following sample(s): \n{active_samples_on_flow_cell}"
                )
                raise DeleteDemuxError

    def delete_flow_cell(
        self,
        cg_stats: bool,
        demultiplexing_dir: bool,
        run_dir: bool,
        housekeeper: bool,
        init_files: bool,
        status_db: bool,
    ) -> None:
        """Master command to delete the presence of a flowcell in all services"""
        self.check_active_samples()
        if status_db:
            self.delete_flow_cell_hasta(
                demultiplexing_dir=True,
                run_dir=True,
            )
            self.delete_flow_cell_cgstats()
            self.delete_flow_cell_housekeeper()
            self.delete_flow_cell_statusdb()
        if demultiplexing_dir or run_dir:
            self.delete_flow_cell_hasta(
                demultiplexing_dir=demultiplexing_dir,
                run_dir=run_dir,
            )
        if cg_stats:
            self.delete_flow_cell_cgstats()
        if init_files and not run_dir:
            self.delete_demux_init_files()
        if housekeeper:
            self.delete_flow_cell_housekeeper()



"""Module to read or write yaml files"""
from pathlib import Path
from typing import Union

import yaml


def read_yaml(file_path: Path) -> Union[dict, list]:
    """Read a yaml into dict"""
    with open(file_path, "r") as file:
        return yaml.safe_load(file)


def read_yaml_stream(stream: str) -> Union[dict, list]:
    "Read yaml formatted stream"
    return yaml.safe_load(stream)


def write_yaml(content: Union[dict, list], file_path: Path) -> None:
    """Write content to a yaml file"""
    with open(file_path, "w") as file:
        yaml.dump(content, file, explicit_start=True)


def write_yaml_stream(content: Union[dict, list]) -> str:
    """Write content to a yaml stream"""
    return yaml.dump(content)


def is_valid_yaml_file(file_path: Path) -> bool:
    """Returns true if the file was successfully loaded as yaml"""
    try:
        read_yaml(file_path=file_path)
    except yaml.YAMLError:
        return False
    return True



"""Tests for the shipping api"""
import logging
from pathlib import Path
from typing import Dict

import pytest

from cg.apps.shipping import ShippingAPI


def build_expected_output_string(
    api: ShippingAPI, app_name: str = None, app_config: Path = None
) -> str:
    """Create a string that is expected to be logged when deploying with shipping"""
    output = ["Running", "command", api.binary_path, "--host-config", api.host_config]
    if app_name:
        output.extend(["--tool-name", app_name])
    if app_config:
        output.extend(["--app-config", str(app_config)])
    output.append("deploy")
    return " ".join(output)


def test_init_shipping_api(shipping_configs: Dict[str, str]):
    """Test to initialize the shipping api"""
    # GIVEN a config with some information about host file and binary
    configs = shipping_configs

    # WHEN initializing the api
    api = ShippingAPI(config=configs)

    # THEN assert that the base command is as expected
    assert api.process.base_call == [
        configs["binary_path"],
        "--host-config",
        configs["host_config"],
    ]


def test_shipping_api_deploy_tool_name(shipping_api: ShippingAPI, caplog):
    """Test to run deploy with the shipping api with just a tool name"""
    caplog.set_level(logging.DEBUG)
    # GIVEN a shipping api and a dummy tool name
    tool_name = "dummy"

    # WHEN running the deploy method
    shipping_api.deploy(app_name=tool_name)

    # THEN assert that call to deploy was communicated
    output_str = build_expected_output_string(api=shipping_api, app_name=tool_name)
    assert output_str in caplog.text


def test_shipping_api_deploy_app_config(shipping_api: ShippingAPI, caplog):
    """Test to run deploy with the shipping api providing an app config"""
    caplog.set_level(logging.DEBUG)
    # GIVEN a shipping api and a dummy app config
    app_config = Path("path/to/app_config.yml")
    app_name = "myapp"

    # WHEN running the deploy method
    shipping_api.deploy(app_name=app_name, app_config=app_config)

    # THEN assert that call to deploy was communicated
    output_str = build_expected_output_string(api=shipping_api, app_config=app_config)

    assert output_str in caplog.text


def test_shipping_api_deploy_shipping(shipping_api: ShippingAPI, caplog):
    """Test to run deploy with shipping itself"""
    caplog.set_level(logging.DEBUG)
    # GIVEN a shipping api
    # GIVEN a shipping app name
    app_name = "shipping"

    # WHEN running the deploy method shipping method
    shipping_api.deploy(app_name=app_name)

    # THEN assert that call to deploy was communicated
    output_str = build_expected_output_string(api=shipping_api, app_name=app_name)
    assert output_str in caplog.text


def test_shipping_api_deploy_genotype(shipping_api: ShippingAPI, caplog):
    """Test to run deploy with genotype"""
    caplog.set_level(logging.DEBUG)
    # GIVEN a shipping api
    # GIVEN a genotype app name
    app_name = "genotype"

    # WHEN running the deploy method shipping method
    shipping_api.deploy(app_name=app_name)

    # THEN assert that call to deploy was communicated
    output_str = build_expected_output_string(api=shipping_api, app_name=app_name)
    assert output_str in caplog.text



""" Test cg.meta.upload.observations module """

import datetime

from cg.models.observations.observations_input_files import ObservationsInputFiles


class Analysis:
    """Mock Analysis object"""

    def __init__(self, case_obj):
        self.case_obj = case_obj

    @property
    def started_at(self):
        """mock started_at date"""
        return str(datetime.datetime.today())

    @property
    def family(self):
        """mock case"""
        return self.case_obj


def test_upload_observations_get_input(analysis_store, case_id, upload_observations_api):

    # GIVEN an upload_observations_api and a mocked analysis_obj
    case_obj = analysis_store.family(case_id)
    analysis_obj = Analysis(case_obj=case_obj)

    # GIVEN a observation_input_files object

    # WHEN get_input method is used given the analysis_obj as argument
    input_files = upload_observations_api.get_input(analysis_obj)

    # THEN input_files is an input_files object
    assert isinstance(input_files, ObservationsInputFiles)

    # THEN sv vcf should be set
    assert input_files.sv_vcf is not None


def test_upload_observations_get_input_wes(analysis_store, case_id, upload_observations_api_wes):

    # GIVEN an upload_observations_api and a mocked analysis_obj
    case_obj = analysis_store.family(case_id)
    analysis_obj = Analysis(case_obj=case_obj)

    # WHEN get_input method is used given the analysis_obj as argument
    input_files = upload_observations_api_wes.get_input(analysis_obj)

    # THEN input_files is an input_files object
    assert isinstance(input_files, ObservationsInputFiles)

    # THEN sv vcf should not be set
    assert input_files.sv_vcf is None


def test_upload_observations_process(analysis_store, case_id, upload_observations_api):

    # GIVEN an upload_observations_api and a mocked analysis_obj
    case_obj = analysis_store.family(case_id)
    analysis_obj = Analysis(case_obj=case_obj)

    # WHEN trying to upload observations

    upload_observations_api.process(analysis_obj)

    # THEN all samples should have gotten a loqusdb_id
    for link in analysis_obj.family.links:
        assert link.sample.loqusdb_id == "123"



import json
from datetime import datetime

from alchy import ModelBase, make_declarative_base


def json_serial(obj):
    """JSON serializer for objects not serializable by default json code"""
    if isinstance(obj, datetime):
        return obj.isoformat()
    raise TypeError("Type not serializable")


class JsonModel(ModelBase):
    def to_json(self, pretty: bool = False):
        """Serialize to JSON.

        Handle DateTime objects.
        """
        kwargs = dict(indent=4, sort_keys=True) if pretty else dict()
        return json.dumps(self.to_dict(), default=json_serial, **kwargs)


Model = make_declarative_base(Base=JsonModel)



import datetime as dt
from typing import List

from cgmodels.cg.constants import Pipeline

from cg.constants import DataDelivery
from cg.exc import OrderError
from cg.meta.orders.lims import process_lims
from cg.meta.orders.submitter import Submitter
from cg.models.orders.order import OrderIn
from cg.models.orders.samples import MicrobialSample
from cg.store import models


class MicrobialSubmitter(Submitter):
    @staticmethod
    def order_to_status(order: OrderIn) -> dict:
        """Convert order input for microbial samples."""

        sample: MicrobialSample
        status_data = {
            "customer": order.customer,
            "order": order.name,
            "comment": order.comment,
            "data_analysis": order.samples[0].data_analysis,
            "data_delivery": order.samples[0].data_delivery,
            "samples": [
                {
                    "application": sample.application,
                    "comment": sample.comment,
                    "control": sample.control,
                    "name": sample.name,
                    "organism_id": sample.organism,
                    "priority": sample.priority,
                    "reference_genome": sample.reference_genome,
                    "volume": sample.volume,
                }
                for sample in order.samples
            ],
        }
        return status_data

    def submit_order(self, order: OrderIn) -> dict:
        self._fill_in_sample_verified_organism(order.samples)
        # submit samples to LIMS
        project_data, lims_map = process_lims(
            lims_api=self.lims, lims_order=order, new_samples=order.samples
        )
        # prepare order for status database
        status_data = self.order_to_status(order)
        self._fill_in_sample_ids(status_data["samples"], lims_map, id_key="internal_id")

        # submit samples to Status
        samples = self.store_items_in_status(
            customer=status_data["customer"],
            order=status_data["order"],
            ordered=project_data["date"] if project_data else dt.datetime.now(),
            ticket=order.ticket,
            items=status_data["samples"],
            comment=status_data["comment"],
            data_analysis=Pipeline(status_data["data_analysis"]),
            data_delivery=DataDelivery(status_data["data_delivery"]),
        )
        return {"project": project_data, "records": samples}

    def store_items_in_status(
        self,
        comment: str,
        customer: str,
        data_analysis: Pipeline,
        data_delivery: DataDelivery,
        order: str,
        ordered: dt.datetime,
        items: List[dict],
        ticket: int,
    ) -> [models.Sample]:
        """Store microbial samples in the status database"""

        sample_objs = []

        customer_obj = self.status.customer(customer)
        new_samples = []

        with self.status.session.no_autoflush:

            for sample_data in items:
                case_obj = self.status.find_family(customer=customer_obj, name=str(ticket))

                if not case_obj:
                    case_obj = self.status.add_case(
                        data_analysis=data_analysis,
                        data_delivery=data_delivery,
                        name=ticket,
                        panels=None,
                    )
                    case_obj.customer = customer_obj
                    self.status.add_commit(case_obj)

                application_tag = sample_data["application"]
                application_version = self.status.current_application_version(application_tag)
                organism = self.status.organism(sample_data["organism_id"])

                if not organism:
                    organism = self.status.add_organism(
                        internal_id=sample_data["organism_id"],
                        name=sample_data["organism_id"],
                        reference_genome=sample_data["reference_genome"],
                    )
                    self.status.add_commit(organism)

                if comment:
                    case_obj.comment = f"Order comment: {comment}"

                new_sample = self.status.add_sample(
                    application_version=application_version,
                    comment=sample_data["comment"],
                    control=sample_data["control"],
                    customer=customer_obj,
                    internal_id=sample_data.get("internal_id"),
                    name=sample_data["name"],
                    order=order,
                    ordered=ordered,
                    organism=organism,
                    priority=sample_data["priority"],
                    reference_genome=sample_data["reference_genome"],
                    sex="unknown",
                    ticket=ticket,
                )

                priority = new_sample.priority
                sample_objs.append(new_sample)
                self.status.relate_sample(family=case_obj, sample=new_sample, status="unknown")
                new_samples.append(new_sample)

            case_obj.priority = priority
            self.status.add_commit(new_samples)
        return sample_objs

    def _fill_in_sample_verified_organism(self, samples: List[MicrobialSample]):
        for sample in samples:
            organism_id = sample.organism
            reference_genome = sample.reference_genome
            organism = self.status.organism(internal_id=organism_id)
            is_verified = (
                organism and organism.reference_genome == reference_genome and organism.verified
            )
            sample.verified_organism = is_verified



"""Constants related to all things sequencing"""
from cgmodels.cg.constants import StrEnum

sequencer_types = {
    "D00134": "hiseqga",
    "D00410": "hiseqga",
    "D00415": "hiseqga",
    "D00450": "hiseqga",
    "D00456": "hiseqga",
    "D00483": "hiseqga",
    "M03284": "hiseqga",
    "SN1025": "hiseqga",
    "SN7001298": "hiseqga",
    "SN7001301": "hiseqga",
    "ST-E00198": "hiseqx",
    "ST-E00201": "hiseqx",
    "ST-E00214": "hiseqx",
    "ST-E00266": "hiseqx",
    "ST-E00269": "hiseqx",
    "A00187": "novaseq",
    "A00621": "novaseq",
    "A00689": "novaseq",
}


class Sequencers(StrEnum):
    NOVASEQ: str = "novaseq"
    HISEQGA: str = "hiseqga"
    HISEQX: str = "hiseqx"
